from __future__ import annotations

from typing import Any

import httpx
from typing_extensions import TypeGuard

from .types.find import (
    Aircraft,
    Airport,
    Entry,
    FindResult,
    Live,
    Operator,
    Schedule,
)


def is_airport(entry: Entry[Any]) -> TypeGuard[Entry[Airport]]:
    return entry["type"] == "airport"


def is_operator(entry: Entry[Any]) -> TypeGuard[Entry[Operator]]:
    return entry["type"] == "operator"


def is_live(entry: Entry[Any]) -> TypeGuard[Entry[Live]]:
    return entry["type"] == "live"


def is_schedule(entry: Entry[Any]) -> TypeGuard[Entry[Schedule]]:
    return entry["type"] == "schedule"


def is_aircraft(entry: Entry[Any]) -> TypeGuard[Entry[Aircraft]]:
    return entry["type"] == "aircraft"


async def find(query: str) -> None | FindResult:
    """
    General search.
    :param query: Airport, schedule (HKG-CDG), or aircraft.
    """
    async with httpx.AsyncClient() as client:
        request = httpx.Request(
            "GET",
            url="https://www.flightradar24.com/v1/search/web/find",
            params={"query": query, "limit": 50},
        )
        response = await client.send(request)
        if response.status_code != 200:
            return None
        return response.json()  # type: ignore



from __future__ import annotations

import asyncio
from datetime import datetime
from pathlib import Path
from typing import Annotated, Optional

import rich
import typer
from appdirs import user_cache_dir, user_config_dir

from .core import FR24, FlightListService, LiveFeedService, PlaybackService
from .tui.tui import main as tui_main

app = typer.Typer(no_args_is_help=True)


@app.command()
def dirs() -> None:
    """Shows relevant directories"""
    rich.print(f"Config: {user_config_dir('fr24')}")
    rich.print(f" Cache: {user_cache_dir('fr24')}")


@app.command()
def tui() -> None:
    """Starts the TUI"""

    tui_main()


app_auth = typer.Typer()
app.add_typer(
    app_auth,
    name="auth",
    no_args_is_help=True,
    help="Commands for authentication",
)


@app_auth.command()
def show() -> None:
    """Shows authentication status"""

    async def show_() -> None:
        async with FR24() as fr24:
            if fr24.http.auth is None:
                rich.print(
                    "[bold red]You are not authenticated.[/bold red]\n"
                    "Provide credentials in environment variables: either \n"
                    "- fr24_username + fr24_password or \n"
                    "- fr24_subscription_key + fr24_token\n"
                    "Alternatively, copy the example config file to "
                    f"{Path(user_config_dir('fr24')) / 'fr24.conf'}."
                )
            else:
                rich.print("[bold green]Authenticated[/bold green]")
                rich.print(fr24.http.auth)

    asyncio.run(show_())


def get_success_message(
    service: FlightListService | LiveFeedService | PlaybackService,
) -> str:
    num_rows = t.num_rows if (t := service.data.table) is not None else 0
    size = service.data.fp.stat().st_size
    return (
        f"[bold green]Success: Saved {num_rows} rows ({size} bytes) "
        f"to {service.data.fp}.[/bold green]"
    )


@app.command()
def feed(
    timestamp: Annotated[
        Optional[int], typer.Option(help="Start unix timestamp (s, UTC)")
    ] = None,
    time: Annotated[
        Optional[datetime], typer.Option(help="Start datetime (UTC)")
    ] = None,
) -> None:
    """Fetches current livefeed / playback of live feed at a given time"""
    # if time is None and timestamp is None:
    #     raise typer.BadParameter("Either time or timestamp must be set")
    if time is not None and timestamp is not None:
        raise typer.BadParameter("Only one of time and timestamp can be set")
    if time is not None:
        timestamp = int(time.timestamp())

    async def feed_() -> None:
        async with FR24() as fr24:
            lf = fr24.livefeed(timestamp)
            lf.data.add_api_response(await lf.api.fetch())
            lf.data.save_parquet()
            rich.print(get_success_message(lf))

    asyncio.run(feed_())


if __name__ == "__main__":
    app()



from __future__ import annotations

import pandas as pd
from fr24.types.fr24 import Airport as AirportJSON
from fr24.types.fr24 import FlightListAircraftData


class Time:
    def __init__(self, timestamp: None | int | str | pd.Timestamp):
        self.ts = timestamp

    def __format__(self, __format_spec: str) -> str:
        if self.ts is None:
            return ""
        if isinstance(self.ts, int):
            ts = pd.Timestamp(self.ts, unit="s", tz="utc")
        else:
            ts = pd.Timestamp(self.ts, tz="utc")
        return format(ts, __format_spec)


class Airport:
    def __init__(self, airport: AirportJSON | None):
        self.airport = airport

    def __format__(self, __format_spec: str) -> str:
        if self.airport is None:
            return ""
        output = __format_spec
        if code := self.airport.get("code"):
            output = output.replace("%a", code["iata"])
            output = output.replace("%o", code["icao"])
        else:
            return ""
        if name := self.airport.get("name"):
            output.replace("%n", name)
        else:
            output = output.replace("%n", "")
        if position := self.airport.get("position"):
            output = output.replace("%y", position["region"]["city"])
        else:
            output = output.replace("%y", "")
        return output


class Aircraft:
    def __init__(self, aircraft: FlightListAircraftData):
        self.aircraft = aircraft

    def __format__(self, __format_spec: str) -> str:
        if self.aircraft is None:
            return ""
        registration = self.aircraft["registration"]
        hexcode = self.aircraft["hex"]
        model_text = self.aircraft["model"]["text"]
        return (
            __format_spec.replace("%c", self.aircraft["model"]["code"])
            .replace("%r", registration if registration else "")
            .replace("%x", hexcode if hexcode else "")
            .replace("%p", model_text if model_text else "")
        )



from __future__ import annotations

from typing import Literal, TypedDict


class FlightListContext(TypedDict):
    reg: str | None
    flight: str | None


class PlaybackContext(TypedDict):
    flight_id: str


class LiveFeedContext(TypedDict):
    timestamp: int | None
    source: Literal["live", "playback"]
    duration: int | None
    hfreq: int | None



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

ADSB: DataSource
AUS: DataSource
BUSINESS_JETS: Service
CARGO: Service
DESCRIPTOR: _descriptor.FileDescriptor
DRONES: Service
ESTIMATED: DataSource
FAA: DataSource
FLARM: DataSource
GENERAL_AVIATION: Service
GLIDERS: Service
GROUND_VEHICLES: Service
HELICOPTERS: Service
LIGHTER_THAN_AIR: Service
MILITARY_AND_GOVERNMENT: Service
MLAT: DataSource
NON_CATEGORIZED: Service
OTHER_DATA_SOURCE: DataSource
OTHER_SERVICE: Service
PASSENGER: Service
SATELLITE: DataSource
SPIDERTRACKS: DataSource
UAT: DataSource

class FiltersList(_message.Message):
    __slots__ = ["airline_filters_list", "airports_list", "altitude_ranges_list", "birth_year_ranges_list", "callsigns_list", "categories_list", "destinations_list", "origins_list", "radars_list", "regs_list", "speed_ranges_list", "types_list"]
    class AirlineFilter(_message.Message):
        __slots__ = ["icao", "type"]
        class Type(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = []
        ICAO_FIELD_NUMBER: _ClassVar[int]
        OPERATED_AS: FiltersList.AirlineFilter.Type
        PAINTED_AS: FiltersList.AirlineFilter.Type
        TYPE_FIELD_NUMBER: _ClassVar[int]
        icao: str
        type: FiltersList.AirlineFilter.Type
        def __init__(self, icao: _Optional[str] = ..., type: _Optional[_Union[FiltersList.AirlineFilter.Type, str]] = ...) -> None: ...
    class AirportFilter(_message.Message):
        __slots__ = ["country_id", "iata", "type"]
        class Type(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = []
        BOTH: FiltersList.AirportFilter.Type
        COUNTRY_ID_FIELD_NUMBER: _ClassVar[int]
        IATA_FIELD_NUMBER: _ClassVar[int]
        INBOUND: FiltersList.AirportFilter.Type
        OUTBOUND: FiltersList.AirportFilter.Type
        TYPE_FIELD_NUMBER: _ClassVar[int]
        country_id: int
        iata: str
        type: FiltersList.AirportFilter.Type
        def __init__(self, iata: _Optional[str] = ..., country_id: _Optional[int] = ..., type: _Optional[_Union[FiltersList.AirportFilter.Type, str]] = ...) -> None: ...
    class AltitudeRange(_message.Message):
        __slots__ = ["max", "min"]
        MAX_FIELD_NUMBER: _ClassVar[int]
        MIN_FIELD_NUMBER: _ClassVar[int]
        max: int
        min: int
        def __init__(self, min: _Optional[int] = ..., max: _Optional[int] = ...) -> None: ...
    class BirthYearRange(_message.Message):
        __slots__ = ["max", "min"]
        MAX_FIELD_NUMBER: _ClassVar[int]
        MIN_FIELD_NUMBER: _ClassVar[int]
        max: int
        min: int
        def __init__(self, min: _Optional[int] = ..., max: _Optional[int] = ...) -> None: ...
    class ODFilter(_message.Message):
        __slots__ = ["country_id", "iata"]
        COUNTRY_ID_FIELD_NUMBER: _ClassVar[int]
        IATA_FIELD_NUMBER: _ClassVar[int]
        country_id: int
        iata: str
        def __init__(self, iata: _Optional[str] = ..., country_id: _Optional[int] = ...) -> None: ...
    class SpeedRange(_message.Message):
        __slots__ = ["max", "min"]
        MAX_FIELD_NUMBER: _ClassVar[int]
        MIN_FIELD_NUMBER: _ClassVar[int]
        max: int
        min: int
        def __init__(self, min: _Optional[int] = ..., max: _Optional[int] = ...) -> None: ...
    AIRLINE_FILTERS_LIST_FIELD_NUMBER: _ClassVar[int]
    AIRPORTS_LIST_FIELD_NUMBER: _ClassVar[int]
    ALTITUDE_RANGES_LIST_FIELD_NUMBER: _ClassVar[int]
    BIRTH_YEAR_RANGES_LIST_FIELD_NUMBER: _ClassVar[int]
    CALLSIGNS_LIST_FIELD_NUMBER: _ClassVar[int]
    CATEGORIES_LIST_FIELD_NUMBER: _ClassVar[int]
    DESTINATIONS_LIST_FIELD_NUMBER: _ClassVar[int]
    ORIGINS_LIST_FIELD_NUMBER: _ClassVar[int]
    RADARS_LIST_FIELD_NUMBER: _ClassVar[int]
    REGS_LIST_FIELD_NUMBER: _ClassVar[int]
    SPEED_RANGES_LIST_FIELD_NUMBER: _ClassVar[int]
    TYPES_LIST_FIELD_NUMBER: _ClassVar[int]
    airline_filters_list: _containers.RepeatedCompositeFieldContainer[FiltersList.AirlineFilter]
    airports_list: _containers.RepeatedCompositeFieldContainer[FiltersList.AirportFilter]
    altitude_ranges_list: _containers.RepeatedCompositeFieldContainer[FiltersList.AltitudeRange]
    birth_year_ranges_list: _containers.RepeatedCompositeFieldContainer[FiltersList.BirthYearRange]
    callsigns_list: _containers.RepeatedScalarFieldContainer[str]
    categories_list: _containers.RepeatedScalarFieldContainer[Service]
    destinations_list: _containers.RepeatedCompositeFieldContainer[FiltersList.ODFilter]
    origins_list: _containers.RepeatedCompositeFieldContainer[FiltersList.ODFilter]
    radars_list: _containers.RepeatedScalarFieldContainer[str]
    regs_list: _containers.RepeatedScalarFieldContainer[str]
    speed_ranges_list: _containers.RepeatedCompositeFieldContainer[FiltersList.SpeedRange]
    types_list: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, altitude_ranges_list: _Optional[_Iterable[_Union[FiltersList.AltitudeRange, _Mapping]]] = ..., speed_ranges_list: _Optional[_Iterable[_Union[FiltersList.SpeedRange, _Mapping]]] = ..., airline_filters_list: _Optional[_Iterable[_Union[FiltersList.AirlineFilter, _Mapping]]] = ..., callsigns_list: _Optional[_Iterable[str]] = ..., radars_list: _Optional[_Iterable[str]] = ..., regs_list: _Optional[_Iterable[str]] = ..., airports_list: _Optional[_Iterable[_Union[FiltersList.AirportFilter, _Mapping]]] = ..., types_list: _Optional[_Iterable[str]] = ..., birth_year_ranges_list: _Optional[_Iterable[_Union[FiltersList.BirthYearRange, _Mapping]]] = ..., origins_list: _Optional[_Iterable[_Union[FiltersList.ODFilter, _Mapping]]] = ..., destinations_list: _Optional[_Iterable[_Union[FiltersList.ODFilter, _Mapping]]] = ..., categories_list: _Optional[_Iterable[_Union[Service, str]]] = ...) -> None: ...

class LiveFeedPlaybackRequest(_message.Message):
    __slots__ = ["hfreq", "live_feed_request", "prefetch", "timestamp"]
    HFREQ_FIELD_NUMBER: _ClassVar[int]
    LIVE_FEED_REQUEST_FIELD_NUMBER: _ClassVar[int]
    PREFETCH_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    hfreq: int
    live_feed_request: LiveFeedRequest
    prefetch: int
    timestamp: int
    def __init__(self, live_feed_request: _Optional[_Union[LiveFeedRequest, _Mapping]] = ..., timestamp: _Optional[int] = ..., prefetch: _Optional[int] = ..., hfreq: _Optional[int] = ...) -> None: ...

class LiveFeedPlaybackResponse(_message.Message):
    __slots__ = ["live_feed_response"]
    LIVE_FEED_RESPONSE_FIELD_NUMBER: _ClassVar[int]
    live_feed_response: LiveFeedResponse
    def __init__(self, live_feed_response: _Optional[_Union[LiveFeedResponse, _Mapping]] = ...) -> None: ...

class LiveFeedRequest(_message.Message):
    __slots__ = ["bounds", "custom_fleet_id", "field_mask", "filters_list", "highlight_mode", "limit", "maxage", "restriction_mode", "selected_flightid", "settings", "stats"]
    class RestrictionMode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    class Bounds(_message.Message):
        __slots__ = ["east", "north", "south", "west"]
        EAST_FIELD_NUMBER: _ClassVar[int]
        NORTH_FIELD_NUMBER: _ClassVar[int]
        SOUTH_FIELD_NUMBER: _ClassVar[int]
        WEST_FIELD_NUMBER: _ClassVar[int]
        east: float
        north: float
        south: float
        west: float
        def __init__(self, north: _Optional[float] = ..., south: _Optional[float] = ..., west: _Optional[float] = ..., east: _Optional[float] = ...) -> None: ...
    class FieldMask(_message.Message):
        __slots__ = ["field_name"]
        FIELD_NAME_FIELD_NUMBER: _ClassVar[int]
        field_name: _containers.RepeatedScalarFieldContainer[str]
        def __init__(self, field_name: _Optional[_Iterable[str]] = ...) -> None: ...
    class Settings(_message.Message):
        __slots__ = ["only_restricted", "services_list", "sources_list", "traffic_type"]
        class TrafficType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = []
        AIRBORNE_ONLY: LiveFeedRequest.Settings.TrafficType
        ALL: LiveFeedRequest.Settings.TrafficType
        GROUND_ONLY: LiveFeedRequest.Settings.TrafficType
        NONE: LiveFeedRequest.Settings.TrafficType
        ONLY_RESTRICTED_FIELD_NUMBER: _ClassVar[int]
        SERVICES_LIST_FIELD_NUMBER: _ClassVar[int]
        SOURCES_LIST_FIELD_NUMBER: _ClassVar[int]
        TRAFFIC_TYPE_FIELD_NUMBER: _ClassVar[int]
        only_restricted: bool
        services_list: _containers.RepeatedScalarFieldContainer[Service]
        sources_list: _containers.RepeatedScalarFieldContainer[DataSource]
        traffic_type: LiveFeedRequest.Settings.TrafficType
        def __init__(self, sources_list: _Optional[_Iterable[_Union[DataSource, str]]] = ..., services_list: _Optional[_Iterable[_Union[Service, str]]] = ..., traffic_type: _Optional[_Union[LiveFeedRequest.Settings.TrafficType, str]] = ..., only_restricted: bool = ...) -> None: ...
    BOUNDS_FIELD_NUMBER: _ClassVar[int]
    CUSTOM_FLEET_ID_FIELD_NUMBER: _ClassVar[int]
    FIELD_MASK_FIELD_NUMBER: _ClassVar[int]
    FILTERS_LIST_FIELD_NUMBER: _ClassVar[int]
    HIGHLIGHT_MODE_FIELD_NUMBER: _ClassVar[int]
    LIMIT_FIELD_NUMBER: _ClassVar[int]
    MAXAGE_FIELD_NUMBER: _ClassVar[int]
    NOT_VISIBLE: LiveFeedRequest.RestrictionMode
    RESTRICTED_INCLUDED: LiveFeedRequest.RestrictionMode
    RESTRICTED_ONLY: LiveFeedRequest.RestrictionMode
    RESTRICTION_MODE_FIELD_NUMBER: _ClassVar[int]
    SELECTED_FLIGHTID_FIELD_NUMBER: _ClassVar[int]
    SETTINGS_FIELD_NUMBER: _ClassVar[int]
    STATS_FIELD_NUMBER: _ClassVar[int]
    bounds: LiveFeedRequest.Bounds
    custom_fleet_id: str
    field_mask: LiveFeedRequest.FieldMask
    filters_list: FiltersList
    highlight_mode: bool
    limit: int
    maxage: int
    restriction_mode: LiveFeedRequest.RestrictionMode
    selected_flightid: _containers.RepeatedScalarFieldContainer[int]
    settings: LiveFeedRequest.Settings
    stats: bool
    def __init__(self, bounds: _Optional[_Union[LiveFeedRequest.Bounds, _Mapping]] = ..., settings: _Optional[_Union[LiveFeedRequest.Settings, _Mapping]] = ..., filters_list: _Optional[_Union[FiltersList, _Mapping]] = ..., custom_fleet_id: _Optional[str] = ..., highlight_mode: bool = ..., stats: bool = ..., limit: _Optional[int] = ..., maxage: _Optional[int] = ..., restriction_mode: _Optional[_Union[LiveFeedRequest.RestrictionMode, str]] = ..., field_mask: _Optional[_Union[LiveFeedRequest.FieldMask, _Mapping]] = ..., selected_flightid: _Optional[_Iterable[int]] = ...) -> None: ...

class LiveFeedResponse(_message.Message):
    __slots__ = ["flights_list", "selected_flight_info", "stats"]
    class FlightData(_message.Message):
        __slots__ = ["altitude", "callsign", "extra_info", "flightid", "ground_speed", "heading", "icon", "latitude", "longitude", "on_ground", "source", "status", "timestamp"]
        class ExtraInfo(_message.Message):
            __slots__ = ["ac_birthday", "airspace", "country_of_reg", "ems_availability", "ems_info", "flight", "icao_address", "logo_id", "reg", "route", "schedule", "squawk", "type", "vspeed"]
            class EMS(_message.Message):
                __slots__ = ["afms", "agps", "agpsdiff", "amcp", "apflags", "ias", "mach", "oat", "qnh", "rs", "tas", "wind_dir", "wind_speed"]
                AFMS_FIELD_NUMBER: _ClassVar[int]
                AGPSDIFF_FIELD_NUMBER: _ClassVar[int]
                AGPS_FIELD_NUMBER: _ClassVar[int]
                AMCP_FIELD_NUMBER: _ClassVar[int]
                APFLAGS_FIELD_NUMBER: _ClassVar[int]
                IAS_FIELD_NUMBER: _ClassVar[int]
                MACH_FIELD_NUMBER: _ClassVar[int]
                OAT_FIELD_NUMBER: _ClassVar[int]
                QNH_FIELD_NUMBER: _ClassVar[int]
                RS_FIELD_NUMBER: _ClassVar[int]
                TAS_FIELD_NUMBER: _ClassVar[int]
                WIND_DIR_FIELD_NUMBER: _ClassVar[int]
                WIND_SPEED_FIELD_NUMBER: _ClassVar[int]
                afms: int
                agps: int
                agpsdiff: int
                amcp: int
                apflags: int
                ias: int
                mach: int
                oat: int
                qnh: int
                rs: int
                tas: int
                wind_dir: int
                wind_speed: int
                def __init__(self, qnh: _Optional[int] = ..., amcp: _Optional[int] = ..., afms: _Optional[int] = ..., oat: _Optional[int] = ..., ias: _Optional[int] = ..., tas: _Optional[int] = ..., mach: _Optional[int] = ..., agps: _Optional[int] = ..., agpsdiff: _Optional[int] = ..., apflags: _Optional[int] = ..., wind_dir: _Optional[int] = ..., wind_speed: _Optional[int] = ..., rs: _Optional[int] = ...) -> None: ...
            class EMSAvailability(_message.Message):
                __slots__ = ["afms_availability", "agps_availability", "agpsdiff_availability", "amcp_availability", "apflags_availability", "ias_availability", "mach_availability", "oat_availability", "qnh_availability", "rs_availability", "tas_availability", "wind_dir_availability", "wind_speed_availability"]
                AFMS_AVAILABILITY_FIELD_NUMBER: _ClassVar[int]
                AGPSDIFF_AVAILABILITY_FIELD_NUMBER: _ClassVar[int]
                AGPS_AVAILABILITY_FIELD_NUMBER: _ClassVar[int]
                AMCP_AVAILABILITY_FIELD_NUMBER: _ClassVar[int]
                APFLAGS_AVAILABILITY_FIELD_NUMBER: _ClassVar[int]
                IAS_AVAILABILITY_FIELD_NUMBER: _ClassVar[int]
                MACH_AVAILABILITY_FIELD_NUMBER: _ClassVar[int]
                OAT_AVAILABILITY_FIELD_NUMBER: _ClassVar[int]
                QNH_AVAILABILITY_FIELD_NUMBER: _ClassVar[int]
                RS_AVAILABILITY_FIELD_NUMBER: _ClassVar[int]
                TAS_AVAILABILITY_FIELD_NUMBER: _ClassVar[int]
                WIND_DIR_AVAILABILITY_FIELD_NUMBER: _ClassVar[int]
                WIND_SPEED_AVAILABILITY_FIELD_NUMBER: _ClassVar[int]
                afms_availability: bool
                agps_availability: bool
                agpsdiff_availability: bool
                amcp_availability: bool
                apflags_availability: bool
                ias_availability: bool
                mach_availability: bool
                oat_availability: bool
                qnh_availability: bool
                rs_availability: bool
                tas_availability: bool
                wind_dir_availability: bool
                wind_speed_availability: bool
                def __init__(self, qnh_availability: bool = ..., amcp_availability: bool = ..., afms_availability: bool = ..., oat_availability: bool = ..., ias_availability: bool = ..., tas_availability: bool = ..., mach_availability: bool = ..., agps_availability: bool = ..., agpsdiff_availability: bool = ..., apflags_availability: bool = ..., wind_dir_availability: bool = ..., wind_speed_availability: bool = ..., rs_availability: bool = ...) -> None: ...
            class Route(_message.Message):
                __slots__ = ["from_", "to"]
                FROM__FIELD_NUMBER: _ClassVar[int]
                TO_FIELD_NUMBER: _ClassVar[int]
                from_: str
                to: str
                def __init__(self, from_: _Optional[str] = ..., to: _Optional[str] = ...) -> None: ...
            class Schedule(_message.Message):
                __slots__ = ["ata", "atd", "eta", "etd", "sta", "std"]
                ATA_FIELD_NUMBER: _ClassVar[int]
                ATD_FIELD_NUMBER: _ClassVar[int]
                ETA_FIELD_NUMBER: _ClassVar[int]
                ETD_FIELD_NUMBER: _ClassVar[int]
                STA_FIELD_NUMBER: _ClassVar[int]
                STD_FIELD_NUMBER: _ClassVar[int]
                ata: int
                atd: int
                eta: int
                etd: int
                sta: int
                std: int
                def __init__(self, std: _Optional[int] = ..., etd: _Optional[int] = ..., atd: _Optional[int] = ..., sta: _Optional[int] = ..., eta: _Optional[int] = ..., ata: _Optional[int] = ...) -> None: ...
            AC_BIRTHDAY_FIELD_NUMBER: _ClassVar[int]
            AIRSPACE_FIELD_NUMBER: _ClassVar[int]
            COUNTRY_OF_REG_FIELD_NUMBER: _ClassVar[int]
            EMS_AVAILABILITY_FIELD_NUMBER: _ClassVar[int]
            EMS_INFO_FIELD_NUMBER: _ClassVar[int]
            FLIGHT_FIELD_NUMBER: _ClassVar[int]
            ICAO_ADDRESS_FIELD_NUMBER: _ClassVar[int]
            LOGO_ID_FIELD_NUMBER: _ClassVar[int]
            REG_FIELD_NUMBER: _ClassVar[int]
            ROUTE_FIELD_NUMBER: _ClassVar[int]
            SCHEDULE_FIELD_NUMBER: _ClassVar[int]
            SQUAWK_FIELD_NUMBER: _ClassVar[int]
            TYPE_FIELD_NUMBER: _ClassVar[int]
            VSPEED_FIELD_NUMBER: _ClassVar[int]
            ac_birthday: str
            airspace: int
            country_of_reg: int
            ems_availability: LiveFeedResponse.FlightData.ExtraInfo.EMSAvailability
            ems_info: LiveFeedResponse.FlightData.ExtraInfo.EMS
            flight: str
            icao_address: int
            logo_id: int
            reg: str
            route: LiveFeedResponse.FlightData.ExtraInfo.Route
            schedule: LiveFeedResponse.FlightData.ExtraInfo.Schedule
            squawk: int
            type: str
            vspeed: int
            def __init__(self, flight: _Optional[str] = ..., reg: _Optional[str] = ..., route: _Optional[_Union[LiveFeedResponse.FlightData.ExtraInfo.Route, _Mapping]] = ..., type: _Optional[str] = ..., squawk: _Optional[int] = ..., vspeed: _Optional[int] = ..., ac_birthday: _Optional[str] = ..., country_of_reg: _Optional[int] = ..., schedule: _Optional[_Union[LiveFeedResponse.FlightData.ExtraInfo.Schedule, _Mapping]] = ..., logo_id: _Optional[int] = ..., airspace: _Optional[int] = ..., ems_info: _Optional[_Union[LiveFeedResponse.FlightData.ExtraInfo.EMS, _Mapping]] = ..., ems_availability: _Optional[_Union[LiveFeedResponse.FlightData.ExtraInfo.EMSAvailability, _Mapping]] = ..., icao_address: _Optional[int] = ...) -> None: ...
        ALTITUDE_FIELD_NUMBER: _ClassVar[int]
        CALLSIGN_FIELD_NUMBER: _ClassVar[int]
        EXTRA_INFO_FIELD_NUMBER: _ClassVar[int]
        FLIGHTID_FIELD_NUMBER: _ClassVar[int]
        GROUND_SPEED_FIELD_NUMBER: _ClassVar[int]
        HEADING_FIELD_NUMBER: _ClassVar[int]
        ICON_FIELD_NUMBER: _ClassVar[int]
        LATITUDE_FIELD_NUMBER: _ClassVar[int]
        LONGITUDE_FIELD_NUMBER: _ClassVar[int]
        ON_GROUND_FIELD_NUMBER: _ClassVar[int]
        SOURCE_FIELD_NUMBER: _ClassVar[int]
        STATUS_FIELD_NUMBER: _ClassVar[int]
        TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
        altitude: int
        callsign: str
        extra_info: LiveFeedResponse.FlightData.ExtraInfo
        flightid: int
        ground_speed: int
        heading: int
        icon: int
        latitude: float
        longitude: float
        on_ground: bool
        source: DataSource
        status: int
        timestamp: int
        def __init__(self, flightid: _Optional[int] = ..., latitude: _Optional[float] = ..., longitude: _Optional[float] = ..., heading: _Optional[int] = ..., altitude: _Optional[int] = ..., ground_speed: _Optional[int] = ..., icon: _Optional[int] = ..., status: _Optional[int] = ..., timestamp: _Optional[int] = ..., on_ground: bool = ..., callsign: _Optional[str] = ..., source: _Optional[_Union[DataSource, str]] = ..., extra_info: _Optional[_Union[LiveFeedResponse.FlightData.ExtraInfo, _Mapping]] = ...) -> None: ...
    class Statistics(_message.Message):
        __slots__ = ["sources"]
        class SourceKV(_message.Message):
            __slots__ = ["count", "source"]
            COUNT_FIELD_NUMBER: _ClassVar[int]
            SOURCE_FIELD_NUMBER: _ClassVar[int]
            count: int
            source: DataSource
            def __init__(self, source: _Optional[_Union[DataSource, str]] = ..., count: _Optional[int] = ...) -> None: ...
        SOURCES_FIELD_NUMBER: _ClassVar[int]
        sources: _containers.RepeatedCompositeFieldContainer[LiveFeedResponse.Statistics.SourceKV]
        def __init__(self, sources: _Optional[_Iterable[_Union[LiveFeedResponse.Statistics.SourceKV, _Mapping]]] = ...) -> None: ...
    FLIGHTS_LIST_FIELD_NUMBER: _ClassVar[int]
    SELECTED_FLIGHT_INFO_FIELD_NUMBER: _ClassVar[int]
    STATS_FIELD_NUMBER: _ClassVar[int]
    flights_list: _containers.RepeatedCompositeFieldContainer[LiveFeedResponse.FlightData]
    selected_flight_info: _containers.RepeatedCompositeFieldContainer[LiveFeedResponse.FlightData]
    stats: LiveFeedResponse.Statistics
    def __init__(self, flights_list: _Optional[_Iterable[_Union[LiveFeedResponse.FlightData, _Mapping]]] = ..., stats: _Optional[_Union[LiveFeedResponse.Statistics, _Mapping]] = ..., selected_flight_info: _Optional[_Iterable[_Union[LiveFeedResponse.FlightData, _Mapping]]] = ...) -> None: ...

class Service(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class DataSource(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from __future__ import annotations

import configparser
import os
from pathlib import Path

import httpx
from appdirs import user_config_dir

from .common import DEFAULT_HEADERS
from .types.fr24 import Authentication

username = os.environ.get("fr24_username", None)
password = os.environ.get("fr24_password", None)
subscription_key = os.environ.get("fr24_subscription_key", None)
token = os.environ.get("fr24_token", None)

if (config_file := (Path(user_config_dir("fr24")) / "fr24.conf")).exists():
    config = configparser.ConfigParser()
    config.read(config_file.as_posix())

    username = config.get("global", "username", fallback=None)
    password = config.get("global", "password", fallback=None)
    subscription_key = config.get("global", "subscription_key", fallback=None)
    token = config.get("global", "token", fallback=None)


async def login(client: httpx.AsyncClient) -> None | Authentication:
    """
    Create credentials for the user (if any). If the environment variables:

    - `subscription_key` and `token` is set: it returns the pair
    - `username` and `password` is set: it fetches the token and
    subscription key from the API
    - otherwise, it returns `None`
    """
    if username is None or password is None:
        if not subscription_key:
            return None
        return {
            "userData": {
                "subscriptionKey": subscription_key,
                "accessToken": token,
            },
            "message": "using subscriptionKey and/or accessToken.",
        }

    res = await client.post(
        "https://www.flightradar24.com/user/login",
        data={"email": username, "password": password},
        headers=DEFAULT_HEADERS,
    )
    res.raise_for_status()
    return res.json()  # type: ignore
    # json['userData']['accessToken']  => Bearer
    # json['userData']['subscriptionKey']  => token



from __future__ import annotations

import asyncio
import json
import time
from datetime import datetime
from pathlib import Path
from typing import Any, AsyncIterator

import pyarrow as pa
import pyarrow.compute as pc
from appdirs import user_cache_dir
from loguru import logger
from typing_extensions import Self

import pandas as pd

from .base import APIBase, ArrowBase, HTTPClient, ServiceBase
from .common import to_unix_timestamp
from .history import (
    flight_list,
    flight_list_dict,
    playback,
    playback_metadata_dict,
    playback_track_dict,
    playback_track_ems_dict,
)
from .livefeed import livefeed_playback_world_data, livefeed_world_data
from .types.cache import (
    LiveFeedRecord,
    flight_list_schema,
    livefeed_schema,
    playback_track_schema,
)
from .types.core import FlightListContext, LiveFeedContext, PlaybackContext
from .types.fr24 import FlightData, FlightList, Playback


class FR24:
    def __init__(self, cache_dir: str = user_cache_dir("fr24")) -> None:
        """
        Populate http clients cache for each service.

        Check the [cache directory](/usage/cli/#directories) for more details.
        """
        self.http = HTTPClient()
        self.cache_dir = cache_dir

    def flight_list(
        self, *, reg: str | None = None, flight: str | None = None
    ) -> FlightListService:
        """
        Constructs a [flight list service][fr24.core.FlightListService] for the
        given registration or flight number.

        *Related: [fr24.history.flight_list][]*

        Input **either** the registration or the flight number, not both.

        :param reg: Aircraft registration (e.g. `B-HUJ`)
        :param flight: Flight number (e.g. `CX8747`)
        """
        if not ((reg is None) ^ (flight is None)):
            raise ValueError(
                "Either reg or flight must be provided, not both or neither."
            )
        ctx: FlightListContext = {"reg": reg, "flight": flight}
        return FlightListService(self.http, self.cache_dir, ctx)

    def playback(self, flight_id: str | int) -> PlaybackService:
        """
        Constructs a [playback service][fr24.core.PlaybackService] for the
        given flight ID.

        *Related: [fr24.history.playback][]*

        :param flight_id: Hex Flight ID (e.g. `"2d81a27"`, `0x2d81a27`)
        """
        if not isinstance(flight_id, str):
            flight_id = f"{flight_id:x}"
        flight_id = flight_id.lower()
        ctx: PlaybackContext = {"flight_id": flight_id}
        return PlaybackService(self.http, self.cache_dir, ctx)

    def livefeed(
        self,
        timestamp: int | str | datetime | pd.Timestamp | None = None,
        *,
        duration: int | None = None,
        hfreq: int | None = None,
    ) -> LiveFeedService:
        """
        Constructs a [live feed service][fr24.core.LiveFeedService].

        :param timestamp: Unix timestamp (seconds) of the live feed data.
            If `None` the [latest live data][fr24.livefeed.livefeed_world_data]
            will be fetched when `.api.fetch()` is called.
            Otherwise,
            [historical data][fr24.livefeed.livefeed_playback_world_data]
            will be fetched instead.
        :param duration: Prefetch duration (default: `7` seconds). Should only
            be set for historical data.
        :param hfreq: High frequency mode (default: `0`). Should only be set for
            historical data.
        """
        ts = to_unix_timestamp(timestamp)
        if ts is None and (hfreq is not None or duration is not None):
            raise ValueError(
                "`hfreq` and `duration` can only be set for historical data."
            )
        ctx: LiveFeedContext = {
            "timestamp": ts,
            "source": "live" if ts is None else "playback",
            "duration": duration,
            "hfreq": hfreq,
        }
        return LiveFeedService(
            self.http,
            self.cache_dir,
            ctx,
        )

    async def __aenter__(self) -> Self:
        await self.http.__aenter__()
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.http.__aexit__(*args)


class FlightListAPI(APIBase[FlightList, FlightListContext]):
    async def fetch(
        self,
        page: int = 1,
        limit: int = 10,
        timestamp: int | datetime | pd.Timestamp | str | None = "now",
    ) -> FlightList:
        """
        Fetch one page of flight list.

        *Related: [fr24.history.flight_list][]*
        """
        return await flight_list(
            self.http.client,
            self.ctx["reg"],
            self.ctx["flight"],
            page,
            limit,
            timestamp,
            self.http.auth,
        )

    async def fetch_all(
        self,
        page: int = 1,
        limit: int = 10,
        timestamp: int | datetime | pd.Timestamp | str | None = "now",
        delay: int = 1,
    ) -> AsyncIterator[FlightList]:
        """
        Iteratively fetch all pages of the flight list.

        *Related: [fr24.history.flight_list][]*

        :param delay: Delay between requests in seconds
        """
        more = True
        while more:
            fl = await self.fetch(page, limit, timestamp)

            # shouldn't happen, but stop in case we overshot
            if (data := fl["result"]["response"]["data"]) is None:
                break

            yield fl

            # NOTE: for the next request, we have to both:
            # - update the timestamp to the earliest STD in the current batch
            # - increment the page
            # weird, but it's how the API works
            timestamp = min(
                t
                for d in data
                if (t := d["time"]["scheduled"]["departure"]) is not None
            )
            page += 1
            more = fl["result"]["response"]["page"]["more"]
            await asyncio.sleep(delay)


class FlightListArrow(ArrowBase[FlightList, FlightListContext]):
    """Arrow table for flight list data."""

    schema = flight_list_schema

    @property
    def fp(self) -> Path:
        if self.ctx["reg"] is not None:
            return (
                self.base_dir
                / "flight_list"
                / "reg"
                / f"{self.ctx['reg'].upper()}.parquet"
            )
        elif self.ctx["flight"] is not None:
            return (
                self.base_dir
                / "flight_list"
                / "flight"
                / f"{self.ctx['flight'].upper()}.parquet"
            )
        raise ValueError("Both reg and flight are None.")

    @staticmethod
    def _concat_tables(tbl_old: pa.Table, tbl_new: pa.Table) -> pa.Table:
        """
        Combine old and new flights, removing duplicates in the old table.

        This should in turn update all aircraft with the "Estimated" status
        with the new data.
        """
        mask: pa.ChunkedArray = pc.is_in(
            tbl_old["flight_id"], value_set=tbl_new["flight_id"]
        )
        if (dup_count := pc.sum(mask).as_py()) > 0:
            logger.warning(f"overwriting {dup_count} duplicate flight ids.")
        return pa.concat_tables(
            [tbl_old.filter(pc.invert(mask.combine_chunks())), tbl_new]
        )

    def add_api_response(self, data: FlightList) -> Self:
        """
        Parse each [fr24.types.fr24.FlightListItem][] in the API response and
        store it in the table.

        :param data: the return data of [fr24.core.FlightListAPI.fetch][].
        """
        if flights := data["result"]["response"]["data"]:
            self.table = pa.Table.from_pylist(
                [flight_list_dict(f) for f in flights],
                schema=self.schema,
            )
        else:
            logger.warning("No data in the API response.")
        return self


class FlightListService(
    ServiceBase[FlightListAPI, FlightListArrow, FlightListContext]
):
    """A service to handle the flight list API and file operations."""

    def __init__(
        self,
        http: HTTPClient,
        base_dir: str,
        ctx: FlightListContext,
    ) -> None:
        api = FlightListAPI(http, ctx)
        fs = FlightListArrow(base_dir, ctx)
        super().__init__(api, fs, ctx)


class PlaybackAPI(APIBase[Playback, PlaybackContext]):
    async def fetch(
        self, timestamp: int | str | datetime | pd.Timestamp | None = None
    ) -> Playback:
        """
        Fetch the historical track playback data for the given flight.

        *Related: [fr24.history.playback][]*

        :param timestamp: Unix timestamp (seconds) of ATD - optional, but
            it is recommended to include it
        """
        return await playback(
            self.http.client, self.ctx["flight_id"], timestamp, self.http.auth
        )


class PlaybackArrow(ArrowBase[Playback, PlaybackContext]):
    """Arrow table for playback data."""

    schema = playback_track_schema

    @property
    def fp(self) -> Path:
        return self.base_dir / "playback" / f"{self.ctx['flight_id']}.parquet"

    @staticmethod
    def _concat_tables(tbl_old: pa.Table, tbl_new: pa.Table) -> pa.Table:
        raise NotImplementedError(
            "Cannot add data to a non-empty playback table.\n"
            "Use `.clear()` to reset the table first."
        )

    def add_api_response(self, data: Playback) -> Self:
        """
        Parse each [fr24.types.fr24.TrackData][] in the API response and store
        it in the table. Also adds [fr24.types.fr24.FlightData][] into the
        schema's metadata with key `_flight`.

        :param data: the return data of [fr24.core.PlaybackAPI.fetch][].
        """
        self.schema = self.schema.with_metadata(
            {
                "_flight": json.dumps(
                    playback_metadata_dict(
                        data["result"]["response"]["data"]["flight"]
                    )
                ).encode("utf-8")
            }
        )
        self.table = pa.Table.from_pylist(
            [
                {
                    **playback_track_dict(point),
                    "ems": playback_track_ems_dict(point),
                }
                for point in data["result"]["response"]["data"]["flight"][
                    "track"
                ]
            ],
            schema=self.schema,
        )
        return self

    @property
    def metadata(self) -> FlightData | None:
        """Get the flight metadata from the arrow schema."""
        if m := self.schema.metadata.get(b"_flight"):
            return json.loads(m)  # type: ignore[no-any-return]
        return None


class PlaybackService(ServiceBase[PlaybackAPI, PlaybackArrow, PlaybackContext]):
    """A service to handle the playback API and file operations."""

    def __init__(
        self, http: HTTPClient, base_dir: str, ctx: PlaybackContext
    ) -> None:
        api = PlaybackAPI(http, ctx)
        fs = PlaybackArrow(base_dir, ctx)
        super().__init__(api, fs, ctx)


class LiveFeedAPI(APIBase[list[LiveFeedRecord], LiveFeedContext]):
    async def fetch(self) -> list[LiveFeedRecord]:
        """
        Fetch the live feed data.

        Updates `self.ctx.timestamp` to the current time if it is `None`.
        """
        if (ts := self.ctx["timestamp"]) is not None:
            kw = {
                k: v
                for k, v in self.ctx.items()
                if k in ("duration", "hfreq") and v is not None
            }
            return await livefeed_playback_world_data(
                self.http.client,
                ts,
                **kw,  # type: ignore[arg-type]
                auth=self.http.auth,
            )
        resp = await livefeed_world_data(self.http.client, self.http.auth)
        self.ctx["timestamp"] = int(time.time())
        return resp


class LiveFeedArrow(ArrowBase[list[LiveFeedRecord], LiveFeedContext]):
    """Arrow table for live feed data."""

    schema = livefeed_schema

    @property
    def fp(self) -> Path:
        ts = self.ctx["timestamp"]
        if self.ctx["source"] == "live" and ts is None:
            raise ValueError(
                "Cannot determine file path for uninitialised live feed.\n"
                "Call `.api.fetch()` first to get the current timestamp."
            )
        return self.base_dir / "feed" / f"{ts}.parquet"

    @staticmethod
    def _concat_tables(tbl_old: pa.Table, tbl_new: pa.Table) -> pa.Table:
        raise NotImplementedError(
            "Cannot add data to a non-empty live feed table.\n"
            "Use `.clear()` to reset the table first."
        )

    def add_api_response(self, data: list[LiveFeedRecord]) -> Self:
        """
        Parse each [fr24.types.cache.LiveFeedRecord][] in the API response and
        store it in the table.

        :param data: the return data of [fr24.core.LiveFeedAPI.fetch][].
        """
        self.schema = self.schema.with_metadata(
            {"ctx": json.dumps(self.ctx).encode("utf-8")}
        )
        self.table = pa.Table.from_pylist(data, schema=self.schema)
        return self


class LiveFeedService(ServiceBase[LiveFeedAPI, LiveFeedArrow, LiveFeedContext]):
    """A service to handle the live feed API and file operations."""

    def __init__(
        self, http: HTTPClient, base_dir: str, ctx: LiveFeedContext
    ) -> None:
        api = LiveFeedAPI(http, ctx)
        fs = LiveFeedArrow(base_dir, ctx)
        super().__init__(api, fs, ctx)



from __future__ import annotations

import asyncio
import secrets
import struct
from typing import Any

import httpx
from loguru import logger

from .common import DEFAULT_HEADERS_GRPC
from .proto.request_pb2 import (
    LiveFeedPlaybackRequest,
    LiveFeedPlaybackResponse,
    LiveFeedRequest,
    LiveFeedResponse,
)
from .types.cache import LiveFeedRecord
from .types.fr24 import Authentication

# N, S, W, E
lng_bounds = [
    -180,
    -117,
    -110,
    -100,
    -95,
    -90,
    -85,
    -82,
    -79,
    -75,
    -68,
    -30,
    -2,
    1,
    5,
    8,
    11,
    15,
    20,
    30,
    40,
    60,
    100,
    110,
    120,
    140,
    180,
]
world_zones = [
    (90, -90, lng_bounds[i], lng_bounds[i + 1])
    for i in range(len(lng_bounds) - 1)
]


def livefeed_message_create(
    north: float = 50,
    south: float = 40,
    west: float = 0,
    east: float = 10,
    stats: bool = False,
    limit: int = 1500,
    maxage: int = 14400,
    fields: list[str] = [
        "flight",
        "reg",
        "route",
        "type",
    ],
    **kwargs: Any,
) -> LiveFeedRequest:
    """
    Create the LiveFeedRequest protobuf message

    :param north: North latitude
    :param south: South latitude
    :param west: West longitude
    :param east: East longitude
    :param stats: Include stats of the given area
    :param limit: Max number of flights
    :param maxage: Max age since last update, seconds
    :param fields: fields to include - for unauthenticated users, max 4 fields.
        When authenticated, `squawk`, `vspeed`, `airspace`, `logo_id` and `age`
        can be included
    """
    return LiveFeedRequest(
        bounds=LiveFeedRequest.Bounds(
            north=north, south=south, west=west, east=east
        ),
        settings=LiveFeedRequest.Settings(
            sources_list=range(10),  # type: ignore
            services_list=range(12),  # type: ignore
            traffic_type=LiveFeedRequest.Settings.ALL,
            only_restricted=False,
        ),
        field_mask=LiveFeedRequest.FieldMask(field_name=fields),
        highlight_mode=False,
        stats=stats,
        limit=limit,
        maxage=maxage,
        restriction_mode=LiveFeedRequest.NOT_VISIBLE,
        **kwargs,
    )


def livefeed_playback_message_create(
    message: LiveFeedRequest,
    timestamp: int,
    prefetch: int,
    hfreq: int,
) -> LiveFeedPlaybackRequest:
    """
    Create the live feed playback request protobuf message.

    :param timestamp: Start timestamp
    :param prefetch: End timestamp: should be start timestamp + 7 seconds
    :param hfreq: High frequency mode
    """
    return LiveFeedPlaybackRequest(
        live_feed_request=message,
        timestamp=timestamp,
        prefetch=prefetch,
        hfreq=hfreq,
    )


def livefeed_request_create(
    message: LiveFeedRequest,
    auth: None | Authentication = None,
) -> httpx.Request:
    """Construct the POST request with encoded gRPC body."""
    request_s = message.SerializeToString()
    post_data = b"\x00" + struct.pack("!I", len(request_s)) + request_s

    headers = DEFAULT_HEADERS_GRPC.copy()
    headers["fr24-device-id"] = f"web-{secrets.token_urlsafe(32)}"
    if auth is not None and auth["userData"]["accessToken"] is not None:
        headers["authorization"] = f"Bearer {auth['userData']['accessToken']}"

    return httpx.Request(
        "POST",
        "https://data-feed.flightradar24.com/fr24.feed.api.v1.Feed/LiveFeed",
        headers=headers,
        content=post_data,
    )


def livefeed_playback_request_create(
    message: LiveFeedPlaybackRequest,
    auth: None | Authentication = None,
) -> httpx.Request:
    """Constructs the POST request with encoded gRPC body."""
    request_s = message.SerializeToString()
    post_data = b"\x00" + struct.pack("!I", len(request_s)) + request_s

    headers = DEFAULT_HEADERS_GRPC.copy()
    headers["fr24-device-id"] = f"web-{secrets.token_urlsafe(32)}"
    if auth is not None and auth["userData"]["accessToken"] is not None:
        headers["authorization"] = f"Bearer {auth['userData']['accessToken']}"

    return httpx.Request(
        "POST",
        "https://data-feed.flightradar24.com/fr24.feed.api.v1.Feed/Playback",
        headers=headers,
        content=post_data,
    )


async def livefeed_post(
    client: httpx.AsyncClient, request: httpx.Request
) -> bytes:
    """Send the request and extract the raw protobuf message."""
    response = await client.send(request)
    data = response.content
    assert len(data) and data[0] == 0
    data_len = int.from_bytes(data[1:5], byteorder="big")
    return data[5 : 5 + data_len]


def livefeed_response_parse(data: bytes) -> LiveFeedResponse:
    """:param data: raw protobuf message"""
    lfr = LiveFeedResponse()
    lfr.ParseFromString(data)
    return lfr


def livefeed_playback_response_parse(data: bytes) -> LiveFeedResponse:
    """:param data: raw protobuf message"""
    lfr = LiveFeedPlaybackResponse()
    lfr.ParseFromString(data)
    return lfr.live_feed_response


def livefeed_flightdata_dict(
    lfr: LiveFeedResponse.FlightData,
) -> LiveFeedRecord:
    """Convert the protobuf message to a dictionary."""
    return {
        "timestamp": lfr.timestamp,
        "flightid": lfr.flightid,
        "latitude": lfr.latitude,
        "longitude": lfr.longitude,
        "heading": lfr.heading,
        "altitude": lfr.altitude,
        "ground_speed": lfr.ground_speed,
        "vertical_speed": lfr.extra_info.vspeed,
        "on_ground": lfr.on_ground,
        "callsign": lfr.callsign,
        "source": lfr.source,
        "registration": lfr.extra_info.reg,
        "origin": lfr.extra_info.route.from_,
        "destination": lfr.extra_info.route.to,
        "typecode": lfr.extra_info.type,
        "eta": lfr.extra_info.schedule.eta,
    }


# TODO: allow for custom bounds
async def livefeed_world_data(
    client: httpx.AsyncClient, auth: None | Authentication = None
) -> list[LiveFeedRecord]:
    """Retrieve live feed data for the entire world, in chunks."""
    results = await asyncio.gather(
        *[
            livefeed_post(
                client,
                livefeed_request_create(
                    livefeed_message_create(*bounds), auth=auth
                ),
            )
            for bounds in world_zones
        ]
    )
    return [
        livefeed_flightdata_dict(lfr)
        for r in results
        for lfr in livefeed_response_parse(r).flights_list
    ]


async def livefeed_playback_world_data(
    client: httpx.AsyncClient,
    timestamp: int,
    duration: int = 7,
    hfreq: int = 0,
    auth: None | Authentication = None,
) -> list[LiveFeedRecord]:
    """
    Retrieve live feed playback data for the entire world, in chunks.
    Raises an exception if more than half of the requests fail.
    """
    results = await asyncio.gather(
        *[
            livefeed_post(
                client,
                livefeed_playback_request_create(
                    livefeed_playback_message_create(
                        livefeed_message_create(*bounds),
                        timestamp,
                        timestamp + duration,
                        hfreq,
                    ),
                    auth=auth,
                ),
            )
            for bounds in world_zones
        ],
        return_exceptions=True,
    )
    if len(err := [r for r in results if not isinstance(r, bytes)]) > 0:
        logger.warning(f"{len(err)} errors: {err}!")
        if len(err) > len(results) / 2:
            raise Exception("Too many errors!")
    return [
        livefeed_flightdata_dict(lfr)
        for r in results
        if isinstance(r, bytes)
        for lfr in livefeed_playback_response_parse(r).flights_list
    ]



from __future__ import annotations

from datetime import datetime

import pandas as pd

DEFAULT_HEADERS = {
    "User-Agent": "Mozilla/5.0 (X11; Linux x86_64; rv:109.0)"
    "Gecko/20100101 Firefox/116.0",
    "Accept": "text/html,application/xhtml+xml,application/xml;"
    "q=0.9,image/avif,image/webp,*/*;q=0.8",
    "Accept-Language": "en-US,en;q=0.5",
    "Accept-Encoding": "gzip, deflate, br",
    "Origin": "https://www.flightradar24.com",
    "Connection": "keep-alive",
    "Referer": "https://www.flightradar24.com/",
    "Sec-Fetch-Dest": "empty",
    "Sec-Fetch-Mode": "cors",
    "Sec-Fetch-Site": "same-site",
    "TE": "trailers",
}

DEFAULT_HEADERS_GRPC = {
    **DEFAULT_HEADERS,
    "Accept": "*/*",
    "fr24-device-id": "web-00000000000000000000000000000000",
    "x-envoy-retry-grpc-on": "unavailable",
    "Content-Type": "application/grpc-web+proto",
    "X-User-Agent": "grpc-web-javascript/0.1",
    "X-Grpc-Web": "1",
    "DNT": "1",
}


def to_unix_timestamp(
    timestamp: int | datetime | pd.Timestamp | str | None,
) -> int | None:
    """
    Casts timestamp-like object to Unix timestamp.
    """
    if isinstance(timestamp, (str, datetime)):
        return int(pd.Timestamp(timestamp).timestamp())
    if isinstance(timestamp, pd.Timestamp):
        return int(timestamp.timestamp())
    return timestamp



from __future__ import annotations

import secrets
from datetime import datetime
from typing import Literal

import httpx

import pandas as pd

from .common import DEFAULT_HEADERS, to_unix_timestamp
from .types.cache import (
    FlightListRecord,
    PlaybackTrackEMSRecord,
    PlaybackTrackRecord,
)
from .types.fr24 import (
    AirportList,
    AirportRequest,
    Authentication,
    FlightData,
    FlightList,
    FlightListItem,
    FlightListRequest,
    Playback,
    PlaybackRequest,
    TrackData,
)


async def flight_list(
    client: httpx.AsyncClient,
    reg: None | str = None,
    flight: None | str = None,
    page: int = 1,
    limit: int = 10,
    timestamp: int | datetime | pd.Timestamp | str | None = "now",
    auth: None | Authentication = None,
) -> FlightList:
    """
    Fetch metadata/history of flights for a given aircraft or flight number.

    Includes basic information such as status, O/D, scheduled/estimated/real
    times: see [fr24.types.fr24.FlightData][] for more details.

    Use *either* `reg` or `flight` to query.
    To determine if there are more pages to query, check the response
    [.result.response.page.more][fr24.types.fr24.Page.more].

    :param client: HTTPX async client
    :param reg: Aircraft registration (e.g. `B-HUJ`)
    :param flight: Flight number (e.g. `CX8747`)
    :param page: Page number
    :param limit: Number of results per page (max 100)
    :param timestamp: Show flights with ATD before this Unix timestamp
    :param auth: Authentication data
    """
    timestamp = to_unix_timestamp(timestamp)

    if reg is not None:
        key, value = "reg", reg
    elif flight is not None:
        key, value = "flight", flight
    else:
        msg = "One named arguments among `reg` and `flight` is expected"
        raise TypeError(msg)

    device = f"web-{secrets.token_urlsafe(32)}"
    headers = DEFAULT_HEADERS.copy()
    headers["fr24-device-id"] = device
    params: FlightListRequest = {
        "query": value,
        "fetchBy": key,
        "page": page,
        "limit": limit,
    }

    if timestamp is not None:
        params["timestamp"] = timestamp

    if auth is not None and auth["userData"]["subscriptionKey"] is not None:
        params["token"] = auth["userData"]["subscriptionKey"]
    else:
        params["device"] = device

    request = httpx.Request(
        "GET",
        "https://api.flightradar24.com/common/v1/flight/list.json",
        headers=headers,
        params=params,  # type: ignore
    )

    response = await client.send(request)
    response.raise_for_status()
    return response.json()  # type: ignore


async def airport_list(
    client: httpx.AsyncClient,
    airport: str,
    mode: Literal["arrivals"] | Literal["departures"] | Literal["ground"],
    page: int = 1,
    limit: int = 10,
    timestamp: int | datetime | pd.Timestamp | str | None = "now",
    auth: None | Authentication = None,
) -> AirportList:
    """
    Fetch aircraft arriving, departing or on ground at a given airport.

    Returns on ground/scheduled/estimated/real times: see
    [fr24.types.fr24.FlightListItem][] for more details.

    :param client: HTTPX async client
    :param airport: IATA airport code (e.g. `HKG`)
    :param mode: arrivals, departures or on ground aircraft
    :param page: Page number
    :param limit: Number of results per page (max 100)
    :param timestamp: Show flights with STA before this Unix timestamp
    :param auth: Authentication data
    """
    timestamp = to_unix_timestamp(timestamp)

    device = f"web-{secrets.token_urlsafe(32)}"
    headers = DEFAULT_HEADERS.copy()
    headers["fr24-device-id"] = device

    params: AirportRequest = {
        "code": airport,
        "plugin[]": ["schedule"],
        "plugin-setting[schedule][mode]": mode,
        "page": page,
        "limit": limit,
    }

    if timestamp is not None:
        params["plugin-setting[schedule][timestamp]"] = timestamp

    if auth is not None and auth["userData"]["subscriptionKey"] is not None:
        params["token"] = auth["userData"]["subscriptionKey"]
    else:
        params["device"] = device

    request = httpx.Request(
        "GET",
        "https://api.flightradar24.com/common/v1/airport.json",
        headers=headers,
        params=params,  # type: ignore
    )

    response = await client.send(request)
    response.raise_for_status()
    return response.json()  # type: ignore


async def playback(
    client: httpx.AsyncClient,
    flight_id: int | str,
    timestamp: int | str | datetime | pd.Timestamp | None = None,
    auth: None | Authentication = None,
) -> Playback:
    """
    Fetch historical track playback data for a given flight.

    :param client: HTTPX async client
    :param flight_id: fr24 hex flight id
    :param timestamp: Unix timestamp (seconds) of ATD - optional, but
    it is recommended to include it
    :param auth: Authentication data
    """
    timestamp = to_unix_timestamp(timestamp)
    if not isinstance(flight_id, str):
        flight_id = f"{flight_id:x}"

    device = f"web-{secrets.token_urlsafe(32)}"
    headers = DEFAULT_HEADERS.copy()
    headers["fr24-device-id"] = device
    params: PlaybackRequest = {
        "flightId": flight_id,
    }
    if timestamp is not None:
        params["timestamp"] = timestamp
    if auth is not None and auth["userData"]["subscriptionKey"] is not None:
        params["token"] = auth["userData"]["subscriptionKey"]
    else:
        params["device"] = device

    request = httpx.Request(
        "GET",
        "https://api.flightradar24.com/common/v1/flight-playback.json",
        headers=headers,
        params=params,  # type: ignore[arg-type]
    )

    response = await client.send(request)
    response.raise_for_status()
    return response.json()  # type: ignore


def playback_metadata_dict(flight: FlightData) -> dict:  # type: ignore[type-arg]
    """
    Flatten and rename important variables in the flight metadata to a
    dictionary.
    """
    ident = flight["identification"]
    sta = flight["status"]["generic"]
    owner = flight["owner"]
    aircraft = flight["aircraft"]
    airline = flight["airline"]
    origin = flight["airport"]["origin"]
    dest = flight["airport"]["destination"]
    return {
        "flight_id": int(id_, 16)
        if isinstance(id_ := ident["id"], str)
        else id_,
        "callsign": ident["callsign"],
        "flight_number": ident["number"].get("default", None),
        "status_type": sta["status"]["type"],
        "status_text": sta["status"]["text"],
        "status_diverted": sta["status"]["diverted"],
        "status_time": int(time)
        if (time := sta["eventTime"]["utc"]) is not None
        else None,
        "model_code": aircraft["model"]["code"]
        if aircraft is not None
        else None,
        "icao24": int(aircraft["identification"]["modes"], 16)
        if aircraft is not None
        else None,
        "registration": aircraft["identification"]["registration"]
        if aircraft is not None
        else None,
        "owner": owner["code"]["icao"] if owner is not None else None,
        "airline": airline["code"]["icao"] if airline is not None else None,
        "origin": origin["code"]["icao"] if origin is not None else None,
        "destination": dest["code"]["icao"] if dest is not None else None,
        "median_delay": delay_med
        if (delay_med := flight["median"]["delay"]) is not None
        else None,
        "median_time": int(time_med)
        if (time_med := flight["median"]["timestamp"]) is not None
        else None,
    }


def playback_track_dict(point: TrackData) -> PlaybackTrackRecord:
    """
    Flatten and rename each variable in this observation into a new dictionary.
    """
    return {
        "timestamp": point["timestamp"],
        "latitude": point["latitude"],
        "longitude": point["longitude"],
        "altitude": point["altitude"]["feet"],
        "ground_speed": point["speed"]["kts"],
        "vertical_speed": point["verticalSpeed"]["fpm"],
        "heading": point["heading"],
        "squawk": int(point["squawk"], base=8),
    }


def playback_track_ems_dict(point: TrackData) -> PlaybackTrackEMSRecord | None:
    """
    If the Extended Mode-S data is available in this observation,
    flatten and rename each variable to a dictionary. Otherwise, return `None`.
    """
    if e := point["ems"]:
        return {
            "timestamp": e["ts"],
            "ias": e["ias"],
            "mach": e["mach"],
            "mcp": e["mcp"],
            "fms": e["fms"],
            "autopilot": e["autopilot"],
            "oat": e["oat"],
            "track": e["trueTrack"],
            "roll": e["rollAngle"],
            "qnh": e["qnh"],
            "wind_dir": e["windDir"],
            "wind_speed": e["windSpd"],
            "precision": e["precision"],
            "altitude_gps": e["altGPS"],
            "emergency": e["emergencyStatus"],
            "tcas_acas": e["tcasAcasDtatus"],
            "heading": e["heading"],
        }
    return None


# TODO: add ems, metadata.
def playback_df(result: Playback) -> pd.DataFrame:
    """
    Transform each point in the flight track to a pandas DataFrame.
    """
    flight = result["result"]["response"]["data"]["flight"]
    df = pd.DataFrame.from_records(
        playback_track_dict(point) for point in flight["track"]
    )

    df = df.eval(
        """
timestamp = @pd.to_datetime(timestamp, unit='s', utc=True)
    """
    )
    return df


def flight_list_dict(entry: FlightListItem) -> FlightListRecord:
    """
    Flatten a flight entry into dict, converting fr24 hex ids into integers.
    """
    orig = entry["airport"]["origin"]
    dest = entry["airport"]["destination"]
    icao24 = entry["aircraft"]["hex"]
    id_ = entry["identification"]["id"]
    stod = entry["time"]["scheduled"]["departure"]
    etod = entry["time"]["estimated"]["departure"]
    atod = entry["time"]["real"]["departure"]
    stoa = entry["time"]["scheduled"]["arrival"]
    etoa = entry["time"]["estimated"]["arrival"]
    atoa = entry["time"]["real"]["arrival"]
    return {
        "flight_id": int(id_, 16) if id_ is not None else None,
        "number": entry["identification"]["number"]["default"],
        "callsign": entry["identification"]["callsign"],
        "icao24": int(icao24, 16) if icao24 is not None else None,
        "registration": entry["aircraft"]["registration"],
        "typecode": entry["aircraft"]["model"]["code"],
        "origin": orig["code"]["icao"] if orig is not None else None,
        "destination": dest["code"]["icao"] if dest is not None else None,
        "status": entry["status"]["text"],
        "STOD": stod * 1000 if stod is not None else None,
        "ETOD": etod * 1000 if etod is not None else None,
        "ATOD": atod * 1000 if atod is not None else None,
        "STOA": stoa * 1000 if stoa is not None else None,
        "ETOA": etoa * 1000 if etoa is not None else None,
        "ATOA": atoa * 1000 if atoa is not None else None,
    }


def flight_list_df(result: FlightList) -> None | pd.DataFrame:
    """
    Transform each flight entry in the JSON response into a pandas DataFrame.
    """
    list_ = result["result"]["response"]["data"]
    if list_ is None:
        return None
    df = pd.DataFrame.from_records(flight_list_dict(entry) for entry in list_)
    return df.eval(
        """
STOD = @pd.to_datetime(STOD, unit="ms", utc=True)
ETOD = @pd.to_datetime(ETOD, unit="ms", utc=True)
ATOD = @pd.to_datetime(ATOD, unit="ms", utc=True)
STOA = @pd.to_datetime(STOA, unit="ms", utc=True)
ETOA = @pd.to_datetime(ETOA, unit="ms", utc=True)
ATOA = @pd.to_datetime(ATOA, unit="ms", utc=True)
"""
    )



from __future__ import annotations

import asyncio
import json
from pathlib import Path
from typing import Iterator, TypeVar

import httpx
from textual import on
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import ScrollableContainer
from textual.widgets import (
    DataTable,
    Footer,
    Header,
    Input,
    Label,
    Static,
)

import pandas as pd
from fr24.authentication import login
from fr24.find import find, is_schedule
from fr24.history import airport_list, flight_list, playback
from fr24.tui.formatters import Aircraft, Airport, Time
from fr24.tui.widgets import AircraftWidget, AirportWidget, FlightWidget
from fr24.types.fr24 import (
    AirportList,
    Authentication,
    FlightList,
    FlightListItem,
)

T = TypeVar("T")


def flatten(*args: list[T]) -> Iterator[T]:
    for elt in args:
        yield from elt


class SearchBlock(Static):
    def compose(self) -> ComposeResult:
        yield Label("date")
        self.date_input = Input(id="date")
        self.date_input.value = f"{pd.Timestamp('now'):%d %b %y}"
        yield self.date_input
        yield AircraftWidget(name="aircraft")
        yield FlightWidget(name="number")
        yield AirportWidget(id="departure", name="origin")
        yield AirportWidget(id="arrival", name="destination")


# -- Application --


class FR24(App[None]):
    CSS_PATH = "style.tcss"
    BINDINGS = [  # noqa: RUF012
        ("q", "quit", "Quit"),
        ("l", "login", "Log in"),
        ("r", "refresh", "Refresh"),  # TODO
        ("/", "search", "Search"),
        ("s", "save", "Save"),
        ("c", "clear", "Clear"),
        Binding("escape", "escape", show=False),
    ]
    line_info: dict[str, str] = {}  # noqa: RUF012

    def compose(self) -> ComposeResult:
        self.auth: Authentication | None = None
        self.client = httpx.AsyncClient()
        self.search_visible = True
        yield Header()
        yield Footer()
        yield SearchBlock()
        yield ScrollableContainer(DataTable())

    def on_mount(self) -> None:
        self.title = "FlightRadar24"
        table = self.query_one(DataTable)
        table.cursor_type = "row"
        table.add_columns(
            "date",
            "number",
            "callsign",
            "aircraft",
            "from",
            "to",
            "STD",
            "ATD",
            "STA",
            "status",
            "flightid",
        )

    def action_search(self) -> None:
        self.search_visible = not self.search_visible
        if self.search_visible:
            self.query_one(SearchBlock).remove_class("hidden")
        else:
            self.query_one(SearchBlock).add_class("hidden")
            self.query_one(DataTable).focus()

    def action_clear(self) -> None:
        for widget in self.query(Input):
            if widget.id != "date":
                widget.value = ""

    async def action_escape(self) -> None:
        if not self.search_visible:
            await self.action_quit()

        self.search_visible = False
        self.query_one(SearchBlock).add_class("hidden")
        self.query_one(DataTable).focus()

    async def action_login(self) -> None:
        self.auth = await login(self.client)
        if self.auth is not None:
            self.sub_title = f"(authenticated: {self.auth['user']['identity']})"
            self.query_one(Header).add_class("authenticated")
            self.query_one(Footer).add_class("authenticated")

    async def on_data_table_row_selected(
        self, event: DataTable.RowSelected
    ) -> None:
        columns = [c.label.plain for c in event.data_table.columns.values()]
        self.line_info = dict(
            zip(columns, event.data_table.get_row(event.row_key))
        )

    async def action_save(self) -> None:
        if len(self.line_info) == 0:
            return
        date = self.line_info["date"] + " " + self.line_info["STD"]
        result = await playback(
            self.client,
            flight_id=self.line_info["flightid"],
            timestamp=date,
            auth=self.auth,
        )
        filename = f"{self.line_info['flightid']}.json"
        self.notify(f"Saving to {filename}")
        Path(filename).write_text(json.dumps(result, indent=2))

    @on(Input.Submitted)
    async def action_refresh(self) -> None:
        ts_str = self.query_one("#date", Input).value
        ts = pd.Timestamp(ts_str if ts_str else "now", tz="utc")

        aircraft_widget = self.query_one(AircraftWidget)
        if aircraft := aircraft_widget.aircraft_id:
            await self.lookup_aircraft(aircraft, ts=ts)
            return
        number_widget = self.query_one(FlightWidget)
        if number := number_widget.number:
            await self.lookup_number(number, ts=ts)
            return
        departure_widget = self.query_one("#departure", AirportWidget)
        arrival_widget = self.query_one("#arrival", AirportWidget)
        if departure := departure_widget.airport_id:
            if arrival := arrival_widget.airport_id:
                await self.lookup_city_pair(departure, arrival, ts=ts)
                return
            await self.lookup_departure(departure, ts=ts)
            return
        if arrival := arrival_widget.airport_id:
            await self.lookup_arrival(arrival, ts=ts)
            return

    async def lookup_aircraft(self, value: str, ts: str) -> None:
        results: FlightList = await flight_list(
            self.client, reg=value, limit=100, timestamp=ts, auth=self.auth
        )
        self.update_table(results["result"]["response"].get("data", None))

    async def lookup_number(self, value: str, ts: str) -> None:
        results: FlightList = await flight_list(
            self.client, flight=value, limit=100, timestamp=ts, auth=self.auth
        )
        self.update_table(results["result"]["response"].get("data", None))

    async def lookup_city_pair(
        self, departure: str, arrival: str, ts: pd.Timestamp
    ) -> None:
        results = await find(f"{departure}-{arrival}")
        if results is None or results["stats"]["count"]["schedule"] == 0:
            return
        flight_numbers = list(
            sched["detail"]["flight"]
            for sched in results["results"]
            if is_schedule(sched)
        )
        flight_lists: list[FlightList] = []
        for value in flight_numbers:
            try:
                res = await flight_list(
                    self.client,
                    flight=value,
                    limit=10,
                    timestamp=ts,
                    auth=self.auth,
                )
            except httpx.HTTPStatusError as exc:
                if exc.response.status_code == 402:  # payment required
                    await asyncio.sleep(10)
                    res = await flight_list(
                        self.client,
                        flight=value,
                        limit=10,
                        timestamp=ts,
                        auth=self.auth,
                    )
                else:
                    raise exc

            flight_lists.append(res)

            compacted_view = list(
                flatten(
                    *(
                        entry
                        for e in flight_lists
                        if (
                            (entry := e["result"]["response"]["data"])
                            is not None
                        )
                    )
                )
            )

            def by_departure_time(elt: FlightListItem) -> int:
                departure_time = elt["time"]["scheduled"]["departure"]
                return -departure_time if departure_time else 0

            compacted_view = sorted(
                (
                    entry
                    for entry in compacted_view
                    if (sobt := entry["time"]["scheduled"]["departure"])
                    is not None
                    and sobt < ts.timestamp() + 3600 * 24
                ),
                key=by_departure_time,
            )
            self.update_table(compacted_view)

            await asyncio.sleep(2)

    async def lookup_arrival(self, value: str, ts: str) -> None:
        results: AirportList = await airport_list(
            self.client,
            airport=value,
            mode="arrivals",
            limit=100,
            timestamp=ts,
            auth=self.auth,
        )
        s = results["result"]["response"]["airport"]["pluginData"]["schedule"]
        data = s["arrivals"].get("data", None)
        if data is not None:
            self.update_table(
                [  # TODO add airport info from
                    elt["flight"]  # type: ignore
                    for elt in data
                ]
            )

    async def lookup_departure(self, value: str, ts: str) -> None:
        results: AirportList = await airport_list(
            self.client,
            airport=value,
            mode="departures",
            limit=100,
            timestamp=ts,
            auth=self.auth,
        )
        s = results["result"]["response"]["airport"]["pluginData"]["schedule"]
        data = s["departures"].get("data", None)
        if data is not None:
            self.update_table(
                [  # TODO add airport info from
                    elt["flight"]  # type: ignore
                    for elt in data
                ]
            )

    def update_table(self, data: None | list[FlightListItem]) -> None:
        table = self.query_one(DataTable)
        table.clear()
        if data is None:
            return
        table.add_rows(
            [
                (
                    f'{Time(entry["time"]["scheduled"]["departure"]):%d %b %y}',
                    entry["identification"]["number"]["default"],
                    entry["identification"]["callsign"],
                    f'{Aircraft(entry["aircraft"]):%r (%c)}',
                    f'{Airport(entry["airport"]["origin"]):%y (%o)}',
                    f'{Airport(entry["airport"]["destination"]):%y (%o)}',
                    f'{Time(entry["time"]["scheduled"]["departure"]):%H:%MZ}',
                    f'{Time(entry["time"]["real"]["departure"]):%H:%MZ}',
                    f'{Time(entry["time"]["scheduled"]["arrival"]):%H:%MZ}',
                    entry["status"]["text"],
                    entry["identification"]["id"],
                )
                for entry in data
            ]
        )


def main() -> None:
    app = FR24()
    app.run()


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import TypedDict

import pyarrow as pa

# NOTE: Parquet does not support timestamp in seconds:
# https://github.com/apache/parquet-format/blob/master/LogicalTypes.md#timestamp
# using u32 or pa.timestamp("ms") for now
flight_list_schema = pa.schema(
    [
        pa.field("flight_id", pa.uint64()),
        pa.field("number", pa.string()),
        pa.field("callsign", pa.string()),
        pa.field("icao24", pa.uint32()),
        pa.field("registration", pa.string()),
        pa.field("typecode", pa.string()),
        pa.field("origin", pa.string()),
        pa.field("destination", pa.string()),
        pa.field("status", pa.string()),
        pa.field("STOD", pa.timestamp("ms")),
        pa.field("ETOD", pa.timestamp("ms")),
        pa.field("ATOD", pa.timestamp("ms")),
        pa.field("STOA", pa.timestamp("ms")),
        pa.field("ETOA", pa.timestamp("ms")),
        pa.field("ATOA", pa.timestamp("ms")),
    ]
)


class FlightListRecord(TypedDict):
    flight_id: int | None
    number: str
    callsign: str | None
    icao24: int | None
    registration: str | None
    typecode: str
    origin: str | None
    destination: str | None
    status: str
    STOD: int | None
    ETOD: int | None
    ATOD: int | None
    STOA: int | None
    ETOA: int | None
    ATOA: int | None


playback_track_schema = pa.schema(
    [
        pa.field("timestamp", pa.uint32()),
        pa.field("latitude", pa.float32()),
        pa.field("longitude", pa.float32()),
        pa.field("altitude", pa.int32()),
        pa.field("ground_speed", pa.int16()),
        pa.field("vertical_speed", pa.int16()),
        pa.field("heading", pa.int16()),
        pa.field("squawk", pa.uint16()),
        pa.field(
            "ems",
            pa.struct(
                [
                    pa.field("timestamp", pa.uint32()),
                    pa.field("ias", pa.int16()),
                    pa.field("mach", pa.int16()),
                    pa.field("mcp", pa.int32()),
                    pa.field("fms", pa.int32()),
                    pa.field("autopilot", pa.int8()),
                    pa.field("oat", pa.int8()),
                    pa.field("track", pa.float32()),
                    pa.field("roll", pa.float32()),
                    pa.field("qnh", pa.uint16()),
                    pa.field("wind_dir", pa.int16()),
                    pa.field("wind_speed", pa.int16()),
                    pa.field("precision", pa.uint8()),
                    pa.field("altitude_gps", pa.int32()),
                    pa.field("emergency", pa.uint8()),
                    pa.field("tcas_acas", pa.uint8()),
                    pa.field("heading", pa.uint16()),
                ]
            ),
        ),
    ],
)


class PlaybackTrackRecord(TypedDict):
    timestamp: int
    latitude: float
    longitude: float
    altitude: int
    ground_speed: int
    vertical_speed: int
    heading: int
    squawk: int


class PlaybackTrackEMSRecord(TypedDict):
    timestamp: int
    ias: int | None
    mach: float | None
    mcp: int | None
    fms: int | None
    autopilot: int | None
    oat: int | None
    track: float | None
    roll: float | None
    qnh: int | None
    wind_dir: int | None
    wind_speed: int | None
    precision: int | None
    altitude_gps: int | None
    emergency: int | None
    tcas_acas: int | None
    heading: int | None


# NOTE: not using pa.timestamp() to save space
# TODO: altitude only requires 12 bits: divide by 25 and store with u16
livefeed_schema = pa.schema(
    [
        pa.field("timestamp", pa.uint32()),
        pa.field("flightid", pa.uint32()),
        pa.field("latitude", pa.float32()),
        pa.field("longitude", pa.float32()),
        pa.field("heading", pa.uint16()),
        pa.field("altitude", pa.int32()),
        pa.field("ground_speed", pa.int16()),
        pa.field("on_ground", pa.bool_()),
        pa.field("callsign", pa.string()),
        pa.field("source", pa.uint8()),
        pa.field("registration", pa.string()),
        pa.field("origin", pa.string()),
        pa.field("destination", pa.string()),
        pa.field("typecode", pa.string()),
        pa.field("eta", pa.uint32()),
        pa.field("vertical_speed", pa.int16()),  # 64 * 9-bit + 1-bit sign
    ]
)


class LiveFeedRecord(TypedDict):
    timestamp: int
    flightid: int
    latitude: float
    longitude: float
    heading: int
    altitude: int
    ground_speed: int
    vertical_speed: int
    on_ground: bool
    callsign: str
    source: int
    registration: str
    origin: str
    destination: str
    typecode: str
    eta: int



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: request.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rrequest.proto\"\xaf\x07\n\x0b\x46iltersList\x12\x38\n\x14\x61ltitude_ranges_list\x18\x01 \x03(\x0b\x32\x1a.FiltersList.AltitudeRange\x12\x32\n\x11speed_ranges_list\x18\x02 \x03(\x0b\x32\x17.FiltersList.SpeedRange\x12\x38\n\x14\x61irline_filters_list\x18\x03 \x03(\x0b\x32\x1a.FiltersList.AirlineFilter\x12\x16\n\x0e\x63\x61llsigns_list\x18\x04 \x03(\t\x12\x13\n\x0bradars_list\x18\x05 \x03(\t\x12\x11\n\tregs_list\x18\x06 \x03(\t\x12\x31\n\rairports_list\x18\x07 \x03(\x0b\x32\x1a.FiltersList.AirportFilter\x12\x12\n\ntypes_list\x18\t \x03(\t\x12;\n\x16\x62irth_year_ranges_list\x18\n \x03(\x0b\x32\x1b.FiltersList.BirthYearRange\x12+\n\x0corigins_list\x18\x0c \x03(\x0b\x32\x15.FiltersList.ODFilter\x12\x30\n\x11\x64\x65stinations_list\x18\r \x03(\x0b\x32\x15.FiltersList.ODFilter\x12!\n\x0f\x63\x61tegories_list\x18\x0e \x03(\x0e\x32\x08.Service\x1a)\n\rAltitudeRange\x12\x0b\n\x03min\x18\x01 \x01(\x05\x12\x0b\n\x03max\x18\x02 \x01(\x05\x1a&\n\nSpeedRange\x12\x0b\n\x03min\x18\x01 \x01(\x05\x12\x0b\n\x03max\x18\x02 \x01(\x05\x1au\n\rAirlineFilter\x12\x0c\n\x04icao\x18\x01 \x01(\t\x12-\n\x04type\x18\x02 \x01(\x0e\x32\x1f.FiltersList.AirlineFilter.Type\"\'\n\x04Type\x12\x0e\n\nPAINTED_AS\x10\x00\x12\x0f\n\x0bOPERATED_AS\x10\x01\x1a\x8d\x01\n\rAirportFilter\x12\x0c\n\x04iata\x18\x01 \x01(\t\x12\x12\n\ncountry_id\x18\x02 \x01(\x05\x12-\n\x04type\x18\x03 \x01(\x0e\x32\x1f.FiltersList.AirportFilter.Type\"+\n\x04Type\x12\x08\n\x04\x42OTH\x10\x00\x12\x0b\n\x07INBOUND\x10\x01\x12\x0c\n\x08OUTBOUND\x10\x02\x1a*\n\x0e\x42irthYearRange\x12\x0b\n\x03min\x18\x01 \x01(\x05\x12\x0b\n\x03max\x18\x02 \x01(\x05\x1a,\n\x08ODFilter\x12\x0c\n\x04iata\x18\x01 \x01(\t\x12\x12\n\ncountry_id\x18\x02 \x01(\x05\"\xf6\x06\n\x0fLiveFeedRequest\x12\'\n\x06\x62ounds\x18\x01 \x01(\x0b\x32\x17.LiveFeedRequest.Bounds\x12+\n\x08settings\x18\x02 \x01(\x0b\x32\x19.LiveFeedRequest.Settings\x12\"\n\x0c\x66ilters_list\x18\x03 \x01(\x0b\x32\x0c.FiltersList\x12\x17\n\x0f\x63ustom_fleet_id\x18\x04 \x01(\t\x12\x16\n\x0ehighlight_mode\x18\x05 \x01(\x08\x12\x12\n\x05stats\x18\x06 \x01(\x08H\x00\x88\x01\x01\x12\x12\n\x05limit\x18\x07 \x01(\x05H\x01\x88\x01\x01\x12\x13\n\x06maxage\x18\x08 \x01(\x05H\x02\x88\x01\x01\x12?\n\x10restriction_mode\x18\t \x01(\x0e\x32 .LiveFeedRequest.RestrictionModeH\x03\x88\x01\x01\x12.\n\nfield_mask\x18\n \x01(\x0b\x32\x1a.LiveFeedRequest.FieldMask\x12\x19\n\x11selected_flightid\x18\x0b \x03(\x07\x1a\x42\n\x06\x42ounds\x12\r\n\x05north\x18\x01 \x01(\x02\x12\r\n\x05south\x18\x02 \x01(\x02\x12\x0c\n\x04west\x18\x03 \x01(\x02\x12\x0c\n\x04\x65\x61st\x18\x04 \x01(\x02\x1a\x83\x02\n\x08Settings\x12!\n\x0csources_list\x18\x01 \x03(\x0e\x32\x0b.DataSource\x12\x1f\n\rservices_list\x18\x02 \x03(\x0e\x32\x08.Service\x12;\n\x0ctraffic_type\x18\x03 \x01(\x0e\x32%.LiveFeedRequest.Settings.TrafficType\x12\x1c\n\x0fonly_restricted\x18\x04 \x01(\x08H\x00\x88\x01\x01\"D\n\x0bTrafficType\x12\x08\n\x04NONE\x10\x00\x12\x0f\n\x0bGROUND_ONLY\x10\x01\x12\x11\n\rAIRBORNE_ONLY\x10\x02\x12\x07\n\x03\x41LL\x10\x03\x42\x12\n\x10_only_restricted\x1a\x1f\n\tFieldMask\x12\x12\n\nfield_name\x18\x01 \x03(\t\"P\n\x0fRestrictionMode\x12\x0f\n\x0bNOT_VISIBLE\x10\x00\x12\x17\n\x13RESTRICTED_INCLUDED\x10\x01\x12\x13\n\x0fRESTRICTED_ONLY\x10\x02\x42\x08\n\x06_statsB\x08\n\x06_limitB\t\n\x07_maxageB\x13\n\x11_restriction_mode\"\xef\r\n\x10LiveFeedResponse\x12\x32\n\x0c\x66lights_list\x18\x01 \x03(\x0b\x32\x1c.LiveFeedResponse.FlightData\x12+\n\x05stats\x18\x02 \x01(\x0b\x32\x1c.LiveFeedResponse.Statistics\x12:\n\x14selected_flight_info\x18\x03 \x03(\x0b\x32\x1c.LiveFeedResponse.FlightData\x1a\xbf\x0b\n\nFlightData\x12\x10\n\x08\x66lightid\x18\x01 \x01(\x05\x12\x10\n\x08latitude\x18\x02 \x01(\x02\x12\x11\n\tlongitude\x18\x03 \x01(\x02\x12\x0f\n\x07heading\x18\x04 \x01(\x05\x12\x10\n\x08\x61ltitude\x18\x05 \x01(\x05\x12\x14\n\x0cground_speed\x18\x06 \x01(\x05\x12\x0c\n\x04icon\x18\x07 \x01(\x05\x12\x0e\n\x06status\x18\x08 \x01(\x05\x12\x11\n\ttimestamp\x18\t \x01(\x05\x12\x11\n\ton_ground\x18\n \x01(\x08\x12\x10\n\x08\x63\x61llsign\x18\x0b \x01(\t\x12\x1b\n\x06source\x18\x0c \x01(\x0e\x32\x0b.DataSource\x12:\n\nextra_info\x18\r \x01(\x0b\x32&.LiveFeedResponse.FlightData.ExtraInfo\x1a\x91\t\n\tExtraInfo\x12\x0e\n\x06\x66light\x18\x01 \x01(\t\x12\x0b\n\x03reg\x18\x02 \x01(\t\x12;\n\x05route\x18\x03 \x01(\x0b\x32,.LiveFeedResponse.FlightData.ExtraInfo.Route\x12\x0c\n\x04type\x18\x04 \x01(\t\x12\x0e\n\x06squawk\x18\x05 \x01(\x05\x12\x0e\n\x06vspeed\x18\x06 \x01(\x05\x12\x13\n\x0b\x61\x63_birthday\x18\x07 \x01(\t\x12\x16\n\x0e\x63ountry_of_reg\x18\x08 \x01(\x05\x12\x41\n\x08schedule\x18\t \x01(\x0b\x32/.LiveFeedResponse.FlightData.ExtraInfo.Schedule\x12\x0f\n\x07logo_id\x18\n \x01(\x05\x12\x10\n\x08\x61irspace\x18\x0b \x01(\x05\x12<\n\x08\x65ms_info\x18\x0c \x01(\x0b\x32*.LiveFeedResponse.FlightData.ExtraInfo.EMS\x12P\n\x10\x65ms_availability\x18\r \x01(\x0b\x32\x36.LiveFeedResponse.FlightData.ExtraInfo.EMSAvailability\x12\x14\n\x0cicao_address\x18\x0e \x01(\x05\x1a\"\n\x05Route\x12\r\n\x05\x66rom_\x18\x01 \x01(\t\x12\n\n\x02to\x18\x02 \x01(\t\x1aX\n\x08Schedule\x12\x0b\n\x03std\x18\x01 \x01(\x05\x12\x0b\n\x03\x65td\x18\x02 \x01(\x05\x12\x0b\n\x03\x61td\x18\x03 \x01(\x05\x12\x0b\n\x03sta\x18\x04 \x01(\x05\x12\x0b\n\x03\x65ta\x18\x05 \x01(\x05\x12\x0b\n\x03\x61ta\x18\x06 \x01(\x05\x1a\xc6\x01\n\x03\x45MS\x12\x0b\n\x03qnh\x18\x01 \x01(\x05\x12\x0c\n\x04\x61mcp\x18\x02 \x01(\x05\x12\x0c\n\x04\x61\x66ms\x18\x03 \x01(\x05\x12\x0b\n\x03oat\x18\x04 \x01(\x05\x12\x0b\n\x03ias\x18\x05 \x01(\x05\x12\x0b\n\x03tas\x18\x06 \x01(\x05\x12\x0c\n\x04mach\x18\x07 \x01(\x05\x12\x0c\n\x04\x61gps\x18\x08 \x01(\x05\x12\x10\n\x08\x61gpsdiff\x18\t \x01(\x05\x12\x0f\n\x07\x61pflags\x18\n \x01(\x05\x12\x10\n\x08wind_dir\x18\x0b \x01(\x05\x12\x12\n\nwind_speed\x18\x0c \x01(\x05\x12\n\n\x02rs\x18\r \x01(\x05\x1a\xfb\x02\n\x0f\x45MSAvailability\x12\x18\n\x10qnh_availability\x18\x01 \x01(\x08\x12\x19\n\x11\x61mcp_availability\x18\x02 \x01(\x08\x12\x19\n\x11\x61\x66ms_availability\x18\x03 \x01(\x08\x12\x18\n\x10oat_availability\x18\x04 \x01(\x08\x12\x18\n\x10ias_availability\x18\x05 \x01(\x08\x12\x18\n\x10tas_availability\x18\x06 \x01(\x08\x12\x19\n\x11mach_availability\x18\x07 \x01(\x08\x12\x19\n\x11\x61gps_availability\x18\x08 \x01(\x08\x12\x1d\n\x15\x61gpsdiff_availability\x18\t \x01(\x08\x12\x1c\n\x14\x61pflags_availability\x18\n \x01(\x08\x12\x1d\n\x15wind_dir_availability\x18\x0b \x01(\x08\x12\x1f\n\x17wind_speed_availability\x18\x0c \x01(\x08\x12\x17\n\x0frs_availability\x18\r \x01(\x08\x1a|\n\nStatistics\x12\x36\n\x07sources\x18\x01 \x03(\x0b\x32%.LiveFeedResponse.Statistics.SourceKV\x1a\x36\n\x08SourceKV\x12\x1b\n\x06source\x18\x01 \x01(\x0e\x32\x0b.DataSource\x12\r\n\x05\x63ount\x18\x02 \x01(\x05\"\x89\x01\n\x17LiveFeedPlaybackRequest\x12+\n\x11live_feed_request\x18\x01 \x01(\x0b\x32\x10.LiveFeedRequest\x12\x11\n\ttimestamp\x18\x02 \x01(\x05\x12\x10\n\x08prefetch\x18\x03 \x01(\x05\x12\x12\n\x05hfreq\x18\x04 \x01(\x05H\x00\x88\x01\x01\x42\x08\n\x06_hfreq\"I\n\x18LiveFeedPlaybackResponse\x12-\n\x12live_feed_response\x18\x01 \x01(\x0b\x32\x11.LiveFeedResponse*\xe6\x01\n\x07Service\x12\r\n\tPASSENGER\x10\x00\x12\t\n\x05\x43\x41RGO\x10\x01\x12\x1b\n\x17MILITARY_AND_GOVERNMENT\x10\x02\x12\x11\n\rBUSINESS_JETS\x10\x03\x12\x14\n\x10GENERAL_AVIATION\x10\x04\x12\x0f\n\x0bHELICOPTERS\x10\x05\x12\x14\n\x10LIGHTER_THAN_AIR\x10\x06\x12\x0b\n\x07GLIDERS\x10\x07\x12\n\n\x06\x44RONES\x10\x08\x12\x13\n\x0fGROUND_VEHICLES\x10\t\x12\x11\n\rOTHER_SERVICE\x10\n\x12\x13\n\x0fNON_CATEGORIZED\x10\x0b*\x8d\x01\n\nDataSource\x12\x08\n\x04\x41\x44SB\x10\x00\x12\x08\n\x04MLAT\x10\x01\x12\t\n\x05\x46LARM\x10\x02\x12\x07\n\x03\x46\x41\x41\x10\x03\x12\r\n\tESTIMATED\x10\x04\x12\r\n\tSATELLITE\x10\x05\x12\x15\n\x11OTHER_DATA_SOURCE\x10\x06\x12\x07\n\x03UAT\x10\x07\x12\x10\n\x0cSPIDERTRACKS\x10\x08\x12\x07\n\x03\x41US\x10\tb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'request_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SERVICE._serialized_start=3846
  _SERVICE._serialized_end=4076
  _DATASOURCE._serialized_start=4079
  _DATASOURCE._serialized_end=4220
  _FILTERSLIST._serialized_start=18
  _FILTERSLIST._serialized_end=961
  _FILTERSLIST_ALTITUDERANGE._serialized_start=527
  _FILTERSLIST_ALTITUDERANGE._serialized_end=568
  _FILTERSLIST_SPEEDRANGE._serialized_start=570
  _FILTERSLIST_SPEEDRANGE._serialized_end=608
  _FILTERSLIST_AIRLINEFILTER._serialized_start=610
  _FILTERSLIST_AIRLINEFILTER._serialized_end=727
  _FILTERSLIST_AIRLINEFILTER_TYPE._serialized_start=688
  _FILTERSLIST_AIRLINEFILTER_TYPE._serialized_end=727
  _FILTERSLIST_AIRPORTFILTER._serialized_start=730
  _FILTERSLIST_AIRPORTFILTER._serialized_end=871
  _FILTERSLIST_AIRPORTFILTER_TYPE._serialized_start=828
  _FILTERSLIST_AIRPORTFILTER_TYPE._serialized_end=871
  _FILTERSLIST_BIRTHYEARRANGE._serialized_start=873
  _FILTERSLIST_BIRTHYEARRANGE._serialized_end=915
  _FILTERSLIST_ODFILTER._serialized_start=917
  _FILTERSLIST_ODFILTER._serialized_end=961
  _LIVEFEEDREQUEST._serialized_start=964
  _LIVEFEEDREQUEST._serialized_end=1850
  _LIVEFEEDREQUEST_BOUNDS._serialized_start=1355
  _LIVEFEEDREQUEST_BOUNDS._serialized_end=1421
  _LIVEFEEDREQUEST_SETTINGS._serialized_start=1424
  _LIVEFEEDREQUEST_SETTINGS._serialized_end=1683
  _LIVEFEEDREQUEST_SETTINGS_TRAFFICTYPE._serialized_start=1595
  _LIVEFEEDREQUEST_SETTINGS_TRAFFICTYPE._serialized_end=1663
  _LIVEFEEDREQUEST_FIELDMASK._serialized_start=1685
  _LIVEFEEDREQUEST_FIELDMASK._serialized_end=1716
  _LIVEFEEDREQUEST_RESTRICTIONMODE._serialized_start=1718
  _LIVEFEEDREQUEST_RESTRICTIONMODE._serialized_end=1798
  _LIVEFEEDRESPONSE._serialized_start=1853
  _LIVEFEEDRESPONSE._serialized_end=3628
  _LIVEFEEDRESPONSE_FLIGHTDATA._serialized_start=2031
  _LIVEFEEDRESPONSE_FLIGHTDATA._serialized_end=3502
  _LIVEFEEDRESPONSE_FLIGHTDATA_EXTRAINFO._serialized_start=2333
  _LIVEFEEDRESPONSE_FLIGHTDATA_EXTRAINFO._serialized_end=3502
  _LIVEFEEDRESPONSE_FLIGHTDATA_EXTRAINFO_ROUTE._serialized_start=2795
  _LIVEFEEDRESPONSE_FLIGHTDATA_EXTRAINFO_ROUTE._serialized_end=2829
  _LIVEFEEDRESPONSE_FLIGHTDATA_EXTRAINFO_SCHEDULE._serialized_start=2831
  _LIVEFEEDRESPONSE_FLIGHTDATA_EXTRAINFO_SCHEDULE._serialized_end=2919
  _LIVEFEEDRESPONSE_FLIGHTDATA_EXTRAINFO_EMS._serialized_start=2922
  _LIVEFEEDRESPONSE_FLIGHTDATA_EXTRAINFO_EMS._serialized_end=3120
  _LIVEFEEDRESPONSE_FLIGHTDATA_EXTRAINFO_EMSAVAILABILITY._serialized_start=3123
  _LIVEFEEDRESPONSE_FLIGHTDATA_EXTRAINFO_EMSAVAILABILITY._serialized_end=3502
  _LIVEFEEDRESPONSE_STATISTICS._serialized_start=3504
  _LIVEFEEDRESPONSE_STATISTICS._serialized_end=3628
  _LIVEFEEDRESPONSE_STATISTICS_SOURCEKV._serialized_start=3574
  _LIVEFEEDRESPONSE_STATISTICS_SOURCEKV._serialized_end=3628
  _LIVEFEEDPLAYBACKREQUEST._serialized_start=3631
  _LIVEFEEDPLAYBACKREQUEST._serialized_end=3768
  _LIVEFEEDPLAYBACKRESPONSE._serialized_start=3770
  _LIVEFEEDPLAYBACKRESPONSE._serialized_end=3843
# @@protoc_insertion_point(module_scope)



import sys
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    TypedDict,
    TypeVar,
    Union,
)


class Live(TypedDict):
    lat: float
    lon: float
    schd_from: str
    schd_to: str
    ac_type: str
    route: str
    logo: str
    reg: str
    callsign: str
    flight: str
    operator: str


class Airport(TypedDict):
    lat: float
    lon: float
    size: float


class Operator(TypedDict):
    operator_id: int
    iata: str
    logo: str


class Schedule(TypedDict):
    logo: str
    callsign: str
    flight: str
    operator: str


class Aircraft(TypedDict):
    equip: str
    hex: str


T = TypeVar("T", Airport, Operator, Live, Schedule, Aircraft)


if sys.version_info >= (3, 11) or TYPE_CHECKING:

    class Entry(Generic[T], TypedDict):
        id: str
        label: str
        detail: T
        type: Literal["airport", "operator", "live", "schedule", "aircraft"]
        match: Literal["icao", "iata", "begins"]
        name: str

else:

    class Entry(TypedDict):
        id: str
        label: str
        detail: dict[str, Any]
        type: Literal["airport", "operator", "live", "schedule", "aircraft"]
        match: Literal["icao", "iata", "begins"]
        name: str


class StatsEntry(TypedDict):
    all: int
    airport: int
    operator: int
    live: int
    schedule: int
    aircraft: int


class Stats(TypedDict):
    total: StatsEntry
    count: StatsEntry


class FindResult(TypedDict):
    results: list[
        Union[
            Entry[Airport],
            Entry[Operator],
            Entry[Live],
            Entry[Schedule],
            Entry[Aircraft],
        ]
    ]
    stats: Stats



from __future__ import annotations

import asyncio
import re

from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widgets import Input, Label, Static

from fr24.find import find, is_aircraft, is_airport, is_schedule


class AirportWidget(Static):
    UPDATE_TASK: None | asyncio.Task[None] = None

    def compose(self) -> ComposeResult:
        assert self.name is not None
        self.iata = Static("", id="iata")
        self.icao = Static("", id="icao")
        self.fullname = Static("", id="name")
        self.airport_id = ""
        self.input = Input(id=self.name)

        yield Label(self.name)
        yield self.input
        with Horizontal():
            yield self.iata
            yield self.icao
        yield self.fullname

    @on(Input.Changed)
    async def lookup_airport(self) -> None:
        self.app.query_one(AircraftWidget).input.value = ""
        self.app.query_one(FlightWidget).input.value = ""
        input_ = self.query_one(Input)
        if len(value := input_.value) >= 3:
            if AirportWidget.UPDATE_TASK is not None:
                AirportWidget.UPDATE_TASK.cancel()
            AirportWidget.UPDATE_TASK = asyncio.create_task(
                self.update_airport(value)
            )
        else:
            self.iata.update()
            self.icao.update()
            self.fullname.update()
            self.airport_id = ""

    def update_info(self, info: None | dict[str, str] = None) -> None:
        if info is None:
            self.icao.update()
            self.iata.update()
            self.fullname.update()
            self.airport_id = ""
        else:
            self.icao.update(info["icao"])
            self.iata.update(info["iata"])
            self.fullname.update(info["name"])
            self.airport_id = info["iata"]

    async def update_airport(self, value: str) -> None:
        find_results = await find(value)
        if find_results is None:
            return self.update_info()
        candidate = next(
            (elt for elt in find_results["results"] if is_airport(elt)),
            None,
        )
        if candidate is None:
            return self.update_info()
        group = re.match(
            r"(?P<name>.+) \((?P<iata>\w+) / (?P<icao>\w+)\)",
            candidate["label"],
        )
        if group is None:
            return self.update_info()
        info: dict[str, str] = group.groupdict()
        return self.update_info(info)


class AircraftWidget(Static):
    UPDATE_TASK: None | asyncio.Task[None] = None

    def compose(self) -> ComposeResult:
        assert self.name is not None
        self.hex = Static("", id="hex")
        self.reg = Static("", id="reg")
        self.type = Static("", id="type")
        self.aircraft_id = ""
        self.input = Input(id=self.name)

        yield Label(self.name)
        yield self.input
        yield self.reg
        with Horizontal():
            yield self.hex
            yield self.type

    @on(Input.Changed)
    async def lookup_aircraft(self) -> None:
        self.app.query_one(FlightWidget).input.value = ""
        for widget in self.app.query(AirportWidget):
            widget.input.value = ""
        input_ = self.query_one(Input)
        if len(value := input_.value) >= 3:
            if AircraftWidget.UPDATE_TASK is not None:
                AircraftWidget.UPDATE_TASK.cancel()
            AircraftWidget.UPDATE_TASK = asyncio.create_task(
                self.update_aircraft(value)
            )
        else:
            self.update_info()

    def update_info(
        self,
        reg: None | str = None,
        hex: None | str = None,
        typecode: None | str = None,
    ) -> None:
        if reg is None or hex is None or typecode is None:
            self.type.update()
            self.hex.update()
            self.reg.update()
            self.aircraft_id = ""
        else:
            self.type.update(typecode)
            self.hex.update(hex.lower())
            self.reg.update(reg)
            self.aircraft_id = reg

    async def update_aircraft(self, value: str) -> None:
        res = await find(value)
        if res is None:
            return self.update_info()
        candidates = (elt for elt in res["results"] if is_aircraft(elt))
        aircraft = next(candidates, None)
        if aircraft is None:
            return self.update_info()

        return self.update_info(
            aircraft["id"],
            aircraft["detail"]["hex"],
            aircraft["detail"]["equip"],
        )


class FlightWidget(Static):
    UPDATE_TASK: None | asyncio.Task[None] = None

    def compose(self) -> ComposeResult:
        assert self.name is not None
        self.flight = Static("", id="flight")
        self.callsign = Static("", id="callsign")
        self.number = ""
        self.input = Input(id=self.name)

        yield Label(self.name)
        yield self.input
        with Horizontal():
            yield self.flight
            yield self.callsign

        self.input.focus()

    @on(Input.Changed)
    async def lookup_number(self) -> None:
        self.app.query_one(AircraftWidget).input.value = ""
        for widget in self.app.query(AirportWidget):
            widget.input.value = ""
        input_ = self.query_one(Input)
        if len(value := input_.value) >= 3:
            if FlightWidget.UPDATE_TASK is not None:
                FlightWidget.UPDATE_TASK.cancel()
            FlightWidget.UPDATE_TASK = asyncio.create_task(
                self.update_number(value)
            )
        else:
            self.update_info()

    def update_info(
        self,
        number: str | None = None,
        callsign: str | None = None,
    ) -> None:
        if number is None or callsign is None:
            self.flight.update()
            self.callsign.update()
            self.number = ""
        else:
            self.flight.update(number)
            self.callsign.update(callsign)
            self.number = number

    async def update_number(self, value: str) -> None:
        find_results = await find(value)
        if find_results is None:
            return self.update_info()
        candidate = next(
            (elt for elt in find_results["results"] if is_schedule(elt)),
            None,
        )
        if candidate is None:
            return self.update_info()
        return self.update_info(
            candidate["detail"]["flight"], candidate["detail"]["callsign"]
        )



from __future__ import annotations

from typing import Literal, Optional, TypedDict

from typing_extensions import Required


class User(TypedDict):
    id: int
    identity: str
    locale: str


class Features(TypedDict): ...  # useless anyway


class UserData(TypedDict, total=False):
    accessToken: None | str
    accountType: str
    countryCode: None | str
    dateExpires: int
    dateLastLogin: str
    features: Features
    hasConsented: bool
    hasPassword: bool
    idUser: int
    identity: str
    isActive: bool
    isAnonymousAccount: bool
    isLoggedIn: bool
    localeCode: str
    name: None
    oAuth: None
    oAuthType: None
    publicKey: None
    subscriptionKey: None | str
    tokenLogin: str
    typeSource: str


class Authentication(TypedDict, total=False):
    message: str
    msg: str
    response_code: int
    responseCode: int
    status: str
    success: bool
    token: str
    user: User
    features: Features
    userData: Required[UserData]


class APIResult(TypedDict):
    copyright: str
    legalNotice: str


class PlaybackRequest(TypedDict, total=False):
    callback: None
    device: None | str
    flightId: Required[str]
    format: Literal["json"]
    pk: None
    timestamp: int
    token: None | str


class FlightNumber(TypedDict):
    default: str
    alternative: None


class FlightIdentification(TypedDict):
    id: str | int
    number: FlightNumber
    callsign: str


class EventTime:
    utc: int
    local: int


class GenericStatus(TypedDict):
    text: str
    type: str
    color: str
    diverted: None
    eventTime: EventTime


class GenericEventTime(TypedDict):
    utc: int | None
    local: int | None


class StatusGeneric(TypedDict):
    status: GenericStatus
    eventTime: GenericEventTime


class StatusData(TypedDict):
    live: bool
    text: str
    icon: str
    estimated: None
    ambiguous: bool
    generic: StatusGeneric


class AircraftModel(TypedDict):
    code: str
    text: None | str


class AircraftAge(TypedDict):
    availability: bool
    date: str


class AircraftAvailability(TypedDict):
    serialNo: bool
    age: bool


class AircraftIdentification(TypedDict):
    modes: str
    registration: str
    serialNo: None | str
    age: None | AircraftAge
    availability: AircraftAvailability


class AircraftData(TypedDict):
    model: AircraftModel
    identification: AircraftIdentification
    availability: AircraftAvailability


class IATA_ICAO(TypedDict):
    iata: str
    icao: str


class OwnerData(TypedDict):
    name: str
    code: IATA_ICAO


class AirlineData(TypedDict):
    name: str
    code: IATA_ICAO
    short: str


class AirportCode(TypedDict):
    iata: str
    icao: str


class Country(TypedDict):
    name: str
    code: str
    id: int


class Region(TypedDict):
    city: str


class AirportPosition(TypedDict):
    latitude: float
    longitude: float
    country: Country
    region: Region


class Timezone(TypedDict):
    name: str
    offset: int
    abbr: str
    abbrName: str
    isDst: bool


class Airport(TypedDict):
    name: str
    code: AirportCode
    position: AirportPosition
    timezone: Timezone
    real: None


# O/D can be null when status is "unknown"
class AirportPairData(TypedDict):
    origin: None | Airport
    destination: None | Airport
    real: None | str


class Median(TypedDict):
    time: int | None
    delay: int | None
    timestamp: int | None


class Altitude(TypedDict):
    feet: int
    meters: int


class Speed(TypedDict):
    kmh: float
    kts: int
    mph: float


class VerticalSpeed(TypedDict):
    fpm: int
    ms: int


class EMS(TypedDict):
    ts: int
    ias: int | None
    tas: int | None
    mach: int | None
    mcp: int | None
    fms: int | None
    autopilot: None
    oat: int | None
    trueTrack: float | None
    rollAngle: float | None
    qnh: None
    windDir: int | None
    windSpd: int | None
    precision: int | None
    altGPS: int | None
    emergencyStatus: int | None
    tcasAcasDtatus: int | None
    heading: int | None


class TrackData(TypedDict):
    latitude: float
    longitude: float
    altitude: Altitude
    speed: Speed
    verticalSpeed: VerticalSpeed
    heading: int
    squawk: str
    timestamp: int
    ems: None | EMS


class Thumbnail(TypedDict):
    src: str
    link: str
    copyright: str
    source: str


class AircraftImages(TypedDict):
    thumbnails: list[Thumbnail]


class FlightData(TypedDict):
    identification: FlightIdentification
    status: StatusData
    aircraft: AircraftData | None
    owner: OwnerData | None
    airline: AirlineData | None
    airport: AirportPairData
    median: Median
    track: list[TrackData]
    aircraftImages: AircraftImages


class PlaybackData(TypedDict):
    flight: FlightData


class PlaybackResponse(TypedDict):
    timestamp: int
    altitudeFiltered: bool
    data: PlaybackData


class PlaybackResult(TypedDict):
    request: PlaybackRequest
    response: PlaybackResponse


class Playback(TypedDict):
    result: PlaybackResult
    _api: APIResult


class FlightListRequest(TypedDict, total=False):
    callback: None
    device: None | str
    fetchBy: Required[str]
    filterBy: str
    format: Literal["json"]
    limit: Required[int]
    olderThenFlightID: None
    page: int
    pk: None
    query: Required[str]
    timestamp: int
    token: None | str


class Item(TypedDict):
    current: int
    total: None | int
    limit: int


class Page(TypedDict):
    current: int
    more: bool
    total: None | int


class Identification(TypedDict):
    id: str | None
    """icao hex"""
    row: int
    """internal row id"""
    number: FlightNumber
    callsign: str | None
    codeshare: None


class FlightListCountry(TypedDict):
    id: None
    name: None
    alpha2: None
    alpha3: None


class FlightListAircraftData(TypedDict):
    model: AircraftModel
    registration: None | str
    country: None | FlightListCountry
    hex: None | str
    restricted: bool
    serialNo: None | str
    age: None | str
    availability: AircraftAvailability
    onGroundUpdate: None | int
    hoursDiff: None | float
    timeDiff: None | int


class Interval(TypedDict):
    departure: None | int
    arrival: None | int


class TimeOther(TypedDict):
    eta: None | int
    updated: None | int
    duration: None | int


class FlightListTime(TypedDict):
    scheduled: Interval
    real: Interval
    estimated: Interval
    other: TimeOther


class FlightListItem(TypedDict):
    identification: Identification
    status: StatusData
    aircraft: FlightListAircraftData
    owner: OwnerData
    airline: AirlineData
    airport: AirportPairData
    time: FlightListTime


class FlightListResponse(TypedDict):
    item: Item
    page: Page
    timestamp: int
    data: None | list[FlightListItem]


class FlightListResult(TypedDict):
    request: FlightListRequest
    response: FlightListResponse


class FlightList(TypedDict):
    result: FlightListResult
    _api: APIResult


class Schedule(TypedDict):
    mode: Literal["departures", "arrivals"] | None
    timestamp: int


Plugin = Literal[
    "details",
    "runways",
    "schedule",
    "satelliteImage",
    "scheduledRoutesStatistics",
    "weather",
]


class PluginSetting(TypedDict, total=False):
    schedule: Schedule


AirportRequest = TypedDict(
    # We need this syntax because of the plugin-setting parameter
    "AirportRequest",
    {
        "callback": None,
        "code": str,
        "device": Optional[str],
        "fleet": Optional[str],
        "format": Literal["json"],
        "limit": int,
        "page": int,
        "pk": None,
        "plugin": list[Plugin],
        "plugin[]": list[Plugin],
        "plugin-setting": PluginSetting,
        "plugin-setting[schedule][mode]": str,
        "plugin-setting[schedule][timestamp]": int,
        "token": Optional[str],
    },
    total=False,
)


class AirportSchedule(TypedDict):
    arrivals: FlightListResponse
    departures: FlightListResponse


class AirportPluginData(TypedDict):
    schedule: AirportSchedule


class AirportListData(TypedDict):
    pluginData: AirportPluginData


class AirportResponse(TypedDict):
    airport: AirportListData


class AirportResult(TypedDict):
    request: AirportRequest
    response: AirportResponse


class AirportList(TypedDict):
    result: AirportResult
    _api: APIResult



from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Generic, TypeVar

import httpx
import pyarrow as pa
import pyarrow.parquet as pq
from loguru import logger
from typing_extensions import Self

import pandas as pd

from .authentication import login
from .types.fr24 import Authentication


class HTTPClient:
    """
    An HTTPX client for making requests to the API and storing
    authentication data.
    """

    def __init__(self, *, retries: int = 5) -> None:
        transport = httpx.AsyncHTTPTransport(retries=retries)
        self.client = httpx.AsyncClient(transport=transport)
        self.auth: Authentication | None = None

    async def __aenter__(self) -> Self:
        self.auth = await login(self.client)
        return self

    async def __aexit__(self, *args: Any) -> None:
        if self.client is not None:
            await self.client.aclose()


ApiRsp = TypeVar("ApiRsp")
"""Type returned by the API, e.g. [fr24.types.fr24.FlightList][]"""
Ctx = TypeVar("Ctx")
"""Type of the context for the service, a TypedDict"""


class APIBase(Generic[ApiRsp, Ctx], ABC):
    def __init__(self, http: HTTPClient, ctx: Ctx) -> None:
        self.http = http
        self.ctx = ctx

    @abstractmethod
    async def fetch(self, *args: Any, **kwargs: Any) -> ApiRsp:
        """Fetch data from the API."""
        ...


class ArrowBase(Generic[ApiRsp, Ctx], ABC):
    """A base class for handling arrow tables."""

    schema: pa.Schema | None = None

    def __init__(self, base_dir: str, ctx: Ctx) -> None:
        self.base_dir = Path(base_dir)
        self._table: pa.Table | None = None
        self.ctx = ctx

    @property
    @abstractmethod
    def fp(self) -> Path:
        """Path to the parquet file."""
        ...

    @property
    def table(self) -> pa.Table | None:
        return self._table

    @table.setter
    def table(self, table: pa.Table) -> None:
        """Set the arrow table or extend it if it already exists."""
        self._table = (
            table
            if self._table is None
            else self._concat_tables(self._table, table)
        )

    @staticmethod
    def _concat_tables(tbl_old: pa.Table, tbl_new: pa.Table) -> pa.Table:
        """Override this to customize how tables should be concatenated."""
        return pa.concat_tables([tbl_old, tbl_new])

    @abstractmethod
    def add_api_response(self, data: ApiRsp) -> Self:
        """Parse API response and add to the arrow table."""

    def add_parquet(self, fp: Path | None = None) -> Self:
        """
        Read parquet file and add to the arrow table.

        :param fp: Path to the parquet file - if not provided, the default cache
        directory is used.
        """
        out_fp = fp if fp is not None else self.fp

        # pq.read_table apparently doesn't read schema metadata:
        # overwriting self.schema using pq.read_schema instead.
        sch = pq.read_schema(out_fp)
        if sch_diffs := set(sch).difference(set(self.schema or [])):
            logger.warning(f"Parquet schema mismatch:\n{sch_diffs}, will cast.")
            sch = sch.cast(self.schema)  # attempt to cast the schema
        self.schema = sch

        self.table = pq.read_table(out_fp, **self._schema_kwargs)
        num_rows = self._table.num_rows if self._table is not None else 0
        logger.debug(f"Read {out_fp}: {num_rows=}")
        return self

    def save_parquet(self, fp: Path | None = None) -> Self:
        """
        Write the arrow table to a parquet file.
        :param fp: Path to the parquet file - if not provided, the default cache
        directory is used.
        """
        if self._table is None:
            logger.warning("No data to write.")
            return self
        out_fp = fp if fp is not None else self.fp
        out_fp.parent.mkdir(parents=True, exist_ok=True)
        with pq.ParquetWriter(out_fp, **self._schema_kwargs) as writer:
            writer.write_table(self._table)
        logger.debug(f"Saved {self._table.num_rows} rows to {out_fp}")
        return self

    def clear(self) -> None:
        """Clear the arrow table."""
        self._table = None

    @property
    def _schema_kwargs(self) -> dict[str, Any]:
        return {} if self.schema is None else {"schema": self.schema}

    @property
    def df(self) -> pd.DataFrame | None:
        return self._table.to_pandas() if self._table is not None else None


ApiBaseT = TypeVar("ApiBaseT")
"""`APIBase[ApiRsp]`"""
FileBaseT = TypeVar("FileBaseT")
"""`FileBase[ApiRsp]`"""


class ServiceBase(Generic[ApiBaseT, FileBaseT, Ctx]):
    """A service to handle the API and file operations."""

    def __init__(self, api: ApiBaseT, data: FileBaseT, ctx: Ctx) -> None:
        self.api = api
        self.data = data
        self.ctx = ctx
